# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Parser
======

Module which forms the basis of a custom :mod:`argparse` based
command line parser which handles setting configuration values
automatically.
"""

import os
from argparse import (
    SUPPRESS, ArgumentParser, _StoreAction, _StoreConstAction,
    _SubParsersAction, _StoreTrueAction, _StoreFalseAction, _AppendAction,
    _AppendConstAction)
from errno import EEXIST
from functools import partial, wraps
from os.path import isdir, isfile, abspath

from netaddr import AddrFormatError, IPAddress

from pyfarm.core.enums import OS, NUMERIC_TYPES
from pyfarm.core.utility import convert

from pyfarm.agent.logger import getLogger
from pyfarm.agent.config import config
from pyfarm.agent.entrypoints.utility import SYSTEMID_MAX

INFINITE = set(["inf", "infinite", "unlimited"])

logger = getLogger("agent.parser")


def assert_parser(func):
    """
    ensures that the instance argument passed along to the validation
    function contains data we expect
    """
    @wraps(func)
    def run(*args, **kwargs):
        parser = kwargs.get("parser")
        assert parser is not None and isinstance(parser, ArgumentParser)
        return func(*args, **kwargs)
    return run


@assert_parser
def ip(value, parser=None, flag=None):
    """make sure the ip address provided is valid"""
    try:
        IPAddress(value)
    except (ValueError, AddrFormatError):
        parser.error("%s is not a valid ip address for %s" % (value, flag))
    else:
        return value


@assert_parser
def port(value, parser=None, get_uid=None, flag=None):
    """convert and check to make sure the provided port is valid"""
    assert callable(get_uid)
    try:
        value = convert.ston(value)
    except (ValueError, SyntaxError):
        parser.error("%s requires a number" % flag)
    else:
        try:
            low_port = 1 if get_uid() == 0 else 49152
        except AttributeError:
            low_port = 49152

        high_port = 65535

        if low_port > value or value > high_port:
            parser.error(
                "valid port range is %s to %s" % (low_port, high_port))

        return value

@assert_parser
def system_identifier(value, parser=None, flag=None):
    """validates a --systemid value"""
    if value == "auto":
        return value

    try:
        value = convert.ston(value)
    except (ValueError, SyntaxError):
        parser.error(
            "failed to convert value provided to %s to an integer" % flag)
    else:
        if 0 >= value or value > SYSTEMID_MAX:
            parser.error("valid range for %s is 0 to %s" % (flag, SYSTEMID_MAX))

        return value

# Function is not currently tested because uid/gid mapping is system specific,
# may require access to external network resources, and internally is
# covered for the most part by other tests.
# TODO: find a reliable way to test uidgid()
@assert_parser
def uidgid(value=None, flag=None,
           get_id=None, check_id=None, set_id=None,
           parser=None):  # pragma: no cover
        """
        Retrieves and validates the user or group id for a command line flag
        """
        # make sure the partial function is setting
        # the input values
        assert flag is not None
        assert get_id is not None
        assert check_id is not None
        assert set_id is not None

        if set_id is NotImplemented:
            logger.info("%s is ignored on %s" % (flag, OS.title()))
            return

        elif not value:
            return

        # convert the incoming argument to a number or fail
        try:
            value = convert.ston(value)
        except ValueError:
            parser.error("failed to convert %s to a number" % flag)

        # make sure the id actually exists
        try:
            check_id(value)
        except KeyError:
            parser.error(
                "%s %s does not seem to exist" % (flag, value))

        # get the uid/gid of the current process
        # so we can reset it after checking it
        original_id = get_id()

        # Try to set the uid/gid to the value requested, fail
        # otherwise.  We're doing this here because we'll be
        # able to stop the rest of the code from running faster
        # and provide a more useful error message.
        try:
            set_id(value)
        except OSError:
            parser.error(
                "Failed to set %s to %s, please make sure you have "
                "the necessary permissions to perform this action.  "
                "Typically you must be running as root." % (flag, value))

        # set the uid/gid back to the original value since
        # the id change should occur inside the form or right
        # before the agent is started
        try:
            set_id(original_id)
        except OSError:
            parser.error(
                "failed to set %s back to the original value" % flag)

        return value


@assert_parser
def direxists(path, parser=None, flag=None, create=False):
    """checks to make sure the directory exists"""
    if create:
        try:
            os.makedirs(path)
        except OSError as e:
            if e.errno != EEXIST:
                parser.error(
                    "Failed to create directory %s: %s" % (path, e))
        else:
            logger.debug("Created %s", path)

    elif not isdir(path):
        parser.error(
            "%s, path does not exist or is not "
            "a directory: %s" % (flag, path))

    return abspath(path)


@assert_parser
def fileexists(path, parser=None, flag=None):
    """checks to make sure the provided file exists"""
    if not isfile(path):
        parser.error(
            "Path %r, which was provided to %s, does not exist" % (path, flag))

    return abspath(path)

@assert_parser
def number(value, types=None, parser=None, allow_inf=False, min_=1,
           flag=None):
    """convert the given value to a number"""
    if value == "auto":
        return value

    # Internally used
    if isinstance(value, NUMERIC_TYPES):  # pragma: no cover
        return value

    elif value.lower() in INFINITE and allow_inf:
        return float("inf")

    elif value.lower() in INFINITE and not allow_inf:
        parser.error("%s does not allow an infinite value" % flag)

    try:
        value = convert.ston(value, types=types or NUMERIC_TYPES)
        if min_ is not None and value < min_:
            parser.error(
                "%s's value must be greater than %s" % (flag, min_))
        return value

    except SyntaxError:  # could not even parse the string as code
        parser.error(
            "%s failed to convert %s to a number" % (flag, repr(value)))

    except ValueError:  # it's a number, but not the right type
        parser.error(
            "%s, %s is not an instance of %s" % (flag, repr(value), types))


@assert_parser
def enum(value, parser=None, enum=None, flag=None):
    """ensures that ``value`` is a valid entry in ``enum``"""
    assert enum is not None
    value = value.lower()

    if value not in enum:
        parser.error(
            "invalid enum value %s for %s, valid values are %s" % (
                value, flag, list(enum)))

    return value

integer = partial(number, types=int)


class ActionMixin(object):
    """
    A mixin which overrides the ``__init__`` and ``__call__``
    methods on an action so we can:

        * Setup attributes to manipulate the config object when the
          arguments are parsed
        * Ensure we all required arguments are present
        * Convert the ``type`` keyword into an internal representation
          so we don't require as much work when we add arguments to
          the parser
    """
    # Maps standard Python functions to some more advanced internal
    # functions.  Our internal versions have several additional
    # options and better error handling.
    TYPE_MAPPING = {
        int: integer,
        isdir: direxists,
        isfile: fileexists}

    def __init__(self, *args, **kwargs):
        self.parser = kwargs.pop("parser")
        self.config = kwargs.pop("config", False)
        type_ = kwargs.get("type")
        type_kwargs = kwargs.pop("type_kwargs", {})

        if self.config is not False:
            if self.config not in config and "default" not in kwargs:
                raise AssertionError(
                    "Config value `%s` does not exist and no default was "
                    "provided.  Please either setup a default in the config "
                    "or provide a default to the argument parser" % self.config)

            # Update the config with the default if one
            # was provided
            if "default" in kwargs:
                default = kwargs["default"]
                if callable(kwargs["default"]):
                    default = default()
                config[self.config] = default

        if type_ is not None:
            assert self.parser is not None
            type_ = self.TYPE_MAPPING.get(type_, type_)

            partial_kwargs = {"parser": self.parser}
            partial_kwargs.update(type_kwargs)

            if "flag" not in partial_kwargs:
                # Convert one of the option string, preferably the
                # one starting with --, and use that as the flag
                # in the type function
                option_strings = kwargs.get("option_strings", [])
                assert len(option_strings) >= 1
                option = option_strings[0]

                for option_string in option_strings:
                    if option.startswith("--"):
                        option = option_string
                        break

                partial_kwargs.update(flag=option)

            kwargs.update(type=partial(type_, **partial_kwargs))

        super(ActionMixin, self).__init__(*args, **kwargs)

        # Before we run the parser, assert that certain flags are
        # set on each action object.
        if self.dest != SUPPRESS:
            if not self.help:
                raise AssertionError(
                    "`help` keyword missing for %s" % self.option_strings)

            if self.config is None:
                raise AssertionError(
                    "The config keyword is missing for %s.  Please provide one "
                    "or set config to False." % self.option_strings)

    def __call__(self, parser, namespace, values, option_string=None):
        super(ActionMixin, self).__call__(
            parser, namespace, values, option_string=option_string)

        # Set the config value based upon the value which was set
        # on the resulting action.  This will be done when parser_args()
        # is called.
        if self.dest != SUPPRESS and self.config not in (False, None):
            config[self.config] = getattr(namespace, self.dest)


#
# Create some classes which mix the class above and the
# original action so we can set attributes and work with
# the configuration
mix_action = lambda class_: type(class_.__name__, (ActionMixin, class_), {})
StoreAction = mix_action(_StoreAction)
SubParsersAction = mix_action(_SubParsersAction)
StoreConstAction = mix_action(_StoreConstAction)
StoreTrueAction = mix_action(_StoreTrueAction)
StoreFalseAction = mix_action(_StoreFalseAction)
AppendAction = mix_action(_AppendAction)
AppendConstAction = mix_action(_AppendConstAction)


class AgentArgumentParser(ArgumentParser):
    """
    A modified :class:`ArgumentParser` which interfaces with
    the agent's configuration.
    """
    def __init__(self, *args, **kwargs):
        super(AgentArgumentParser, self).__init__(*args, **kwargs)

        # Override the relevant actions with out own so
        # we have more control of what they're doing
        self.register("action", None,
                      partial(StoreAction, parser=self))
        self.register("action", "store",
                      partial(StoreAction, parser=self))
        self.register("action", "store_const",
                      partial(StoreConstAction, parser=self))
        self.register("action", "store_true",
                      partial(StoreTrueAction, parser=self))
        self.register("action", "store_false",
                      partial(StoreFalseAction, parser=self))
        self.register("action", "append",
                      partial(AppendAction, parser=self))
        self.register("action", "append_const",
                      partial(AppendConstAction, parser=self))
        self.register("action", "parsers",
                      partial(SubParsersAction, parser=self))



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
# Copyright 2014 Ambient Entertainment GmbH & Co. KG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Update Endpoint
===============

This endpoint is used to instruct the agent to download and apply an update.
"""

try:
    from httplib import OK, ACCEPTED
except ImportError:  # pragma: no cover
    from http.client import OK, ACCEPTED

from os import makedirs
from os.path import join, isdir

from treq import get, collect
from twisted.internet import reactor
from twisted.web.server import NOT_DONE_YET
import twisted.python.failure

from voluptuous import Schema, Required

from pyfarm.agent.http.api.base import APIResource
from pyfarm.agent.http.core.client import http_retry_delay
from pyfarm.agent.config import config
from pyfarm.agent.logger import getLogger
from pyfarm.agent.utility import STRINGS

logger = getLogger("agent.http.update")


class Update(APIResource):
    """
    Requests the agent to download and apply the specified version of itself.
    Will make the agent restart at the next opportunity.

    .. http:post:: /api/v1/update HTTP/1.1

        **Request**

        .. sourcecode:: http

            POST /api/v1/update HTTP/1.1
            Accept: application/json

            {
                "version": 1.2.3
            }

        **Response**

        .. sourcecode:: http

            HTTP/1.1 200 ACCEPTED
            Content-Type: application/json

    """
    SCHEMAS = {
        "POST": Schema({
            Required("version"): STRINGS})}
    isLeaf = False  # this is not really a collection of things

    def post(self, **kwargs):
        request = kwargs["request"]
        data = kwargs["data"]
        agent = config["agent"]

        # TODO Check version parameter for sanity

        url = "%s/agents/updates/%s" % (config["master_api"], data["version"])
        url = url.encode()

        def download_update(version):
            config["downloading_update"] = True
            deferred = get(url)
            deferred.addCallback(update_downloaded)
            deferred.addErrback(update_failed)

        def update_failed(arg):
            logger.error("Downloading update failed: %r", arg)
            if isinstance(arg, twisted.python.failure.Failure):
                logger.error("Traceback: %s", arg.getTraceback())
            reactor.callLater(http_retry_delay(),
                              download_update,
                              data["version"])

        def update_downloaded(response):
            if response.code == OK:
                update_filename = join(config["agent_updates_dir"],
                                       "pyfarm-agent.zip")
                logger.debug("Writing update to %s", update_filename)
                if not isdir(config["agent_updates_dir"]):
                    makedirs(config["agent_updates_dir"])
                with open(update_filename, "wb+") as update_file:
                    collect(response, update_file.write)
                logger.info("Update file for version %s has been downloaded "
                            "and stored under %s", data["version"],
                            update_filename)
                # TODO Only shut down if we were started by the supervisor
                config["restart_requested"] = True
                if len(config["current_assignments"]) == 0:
                    agent.stop()
            else:
                logger.error("Unexpected return code %s on downloading update "
                             "from master", response.code)

            config["downloading_update"] = False

        if ("downloading_update" not in config or
            not config["downloading_update"]):
            download_update(data["version"])

        request.setResponseCode(ACCEPTED)
        request.finish()

        return NOT_DONE_YET



# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
System
------

Information about the operating system including type, filesystem information,
and other relevant information.  This module may also contain os specific
information such as the Linux distribution, Windows version, bitness, etc.
"""

import os
import platform
import sys
import time
import tempfile
import uuid
from os.path import isfile

import psutil

try:
    WindowsError
except NameError:  # pragma: no cover
    WindowsError = OSError

from pyfarm.agent.logger import getLogger
from pyfarm.agent.sysinfo.network import mac_addresses

logger = getLogger("agent.sysinfo")


def filesystem_is_case_sensitive():  # pragma: no cover
    """returns True if the file system is case sensitive"""
    fd, path = tempfile.mkstemp()
    case_sensitive = not all(map(isfile, [path, path.lower(), path.upper()]))

    try:
        os.remove(path)
    except (WindowsError, OSError, NotImplementedError):
        pass

    return case_sensitive


def environment_is_case_sensitive():
    """returns True if the environment is case sensitive"""
    envvar_lower = "PYFARM_CHECK_ENV_CASE_" + uuid.uuid4().hex
    envvar_upper = envvar_lower.upper()

    # populate environment then compare the difference
    os.environ.update({envvar_lower: "0", envvar_upper: "1"})
    case_sensitive = os.environ[envvar_lower] != os.environ[envvar_upper]

    # remove the envvars we just made
    for envvar in (envvar_lower, envvar_upper):
        os.environ.pop(envvar, None)

    return case_sensitive


def machine_architecture(arch=platform.machine().lower()):
    """returns the architecture of the host itself"""
    if arch in ("amd64", "x86_64", "ia64") or "wow64" in arch:
        return 64

    elif arch in ("i386", "i686", "x86"):
        return 32

    elif not arch:
        raise NotImplementedError(
            "Cannot handle `arch` being unpopulated.")

    else:
        raise NotImplementedError(
            "Don't know how to handle a machine architecture %s" % repr(arch))


# Don't collect coverage because it's using the internal Python
# implementation which varies between platforms.
def interpreter_architecture():  # pragma: no cover
    """returns the architecture of the interpreter itself (32 or 64)"""
    if hasattr(sys, "maxsize"):
        if sys.maxsize > 2**32:
            return 64
        else:
            return 32
    else:
        # Python < 2.6, not as accurate as the above
        if platform.architecture()[0] == "64bits":
            return 64
        else:
            return 32


def uptime():
    """
    Returns the amount of time the system has been running in
    seconds.
    """
    return time.time() - psutil.boot_time()


def operating_system(plat=sys.platform):
    """
    Returns the operating system for the given platform.  Please
    note that while you can call this function directly you're more
    likely better off using values in :mod:`pyfarm.core.enums` instead.
    """
    if plat.startswith("linux"):
        return "linux"
    elif plat.startswith("win"):
        return "windows"
    elif plat.startswith("darwin"):
        return "mac"
    else:
        logger.warning("unknown operating system: %r", plat)
        return "other"


def system_identifier():
    """Generates a system identifier"""
    result = 0
    for address in mac_addresses(long_addresses=False, as_integers=True):
        result ^= address
    return result



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Configuration
-------------

Central module for storing and working with a live configuration objects.  This
module instances :class:`.ConfigurationWithCallbacks` onto :const:`.config`.
Attempting to reload this module will not reinstance the :const:`.config`
object.

The :const:`.config` object should be directly imported from this
module to be used:

    >>> from pyfarm.agent.config import config
"""

import os
from datetime import datetime
from os.path import join, abspath, dirname

from pyfarm.core.enums import NOTSET
from pyfarm.core.config import Configuration
from pyfarm.agent.logger import getLogger
from pyfarm.agent.sysinfo import memory, cpu, network, system

logger = getLogger("agent.config")


class LoggingConfiguration(Configuration):
    """
    Special configuration object which logs when a key is changed in
    a dictionary.  If the reactor is not running then log messages will
    be queued until they can be emitted so they are not lost.

    .. automethod:: _expandvars
    """
    MODIFIED = "modified"
    CREATED = "created"
    DELETED = "deleted"

    def __init__(self, data=None, environment=None, load=True):
        super(LoggingConfiguration, self).__init__("pyfarm.agent")
        assert data is None or isinstance(data, dict)
        assert environment is None or isinstance(environment, dict)

        if environment is None:
            environment = os.environ

        if load:
            self.load(environment=environment)
            self.update(
                # A mapping of UUIDs to job type instances.
                jobtypes={},

                # A mapping of tasks to job type instances.
                current_assignments={},

                # The last time we were in touch with the master,
                # or the last time it was in touch with us.
                last_master_contact=None,

                # The last time we announced ourselves to the master.  This
                # may be longer than --master-reannounce if
                # `last_master_contact` caused us to skip an announcement.
                last_announce=None)

        if data is not None:
            self.update(data)

        # Load configuration file(s) for jobtypes and then
        # update the local instance
        if load:
            jobtypes_config = Configuration(
                "pyfarm.jobtypes", version=self.version)
            jobtypes_config.load(environment=environment)
            self.update(jobtypes_config)

    def _map_value(self, key, value):
        """
        Some configuration values have keywords associated with
        them, this function is responsible for returning the 'fixed'
        value.
        """
        if value == "auto":
            if key == "agent_ram":
                return memory.total_ram()

            if key == "agent_cpus":
                return cpu.total_cpus()

            if key == "agent_hostname":
                return network.hostname()

            if key == "agent_static_root":
                return abspath(
                    join(dirname(__file__), "http", "static"))

        return value

    def __setitem__(self, key, value):
        value = self._map_value(key, value)

        if key not in self:
            self.changed(self.CREATED, key, value, NOTSET)
        elif self[key] != value:
            self.changed(self.MODIFIED, key, value, self[key])

        # Run the base class's method after the above otherwise
        # the value would already be in the data we're comparing
        # against
        super(LoggingConfiguration, self).__setitem__(key, value)

    def __delitem__(self, key):
        """
        Deletes the provided ``key`` and triggers a ``delete`` event
        using :meth:`.changed`.
        """
        old_value = self[key] if key in self else NOTSET
        super(LoggingConfiguration, self).__delitem__(key)
        self.changed(self.DELETED, key, NOTSET, old_value)

    def pop(self, key, *args):
        """
        Deletes the provided ``key`` and triggers a ``delete`` event
        using :meth:`.changed`.
        """
        old_value = self[key] if key in self else NOTSET
        super(LoggingConfiguration, self).pop(key, *args)
        self.changed(self.DELETED, key, NOTSET, old_value)

    def clear(self):
        """
        Deletes all keys in this object and triggers a ``delete`` event
        using :meth:`.changed` for each one.
        """
        keys = list(self.keys())

        # Not quite the same thing as dict.clear() but the effect
        # is the same as the call to changed() is more real time.
        for key in keys:
            old_value = self.pop(key, NOTSET)
            self.changed(self.DELETED, key, NOTSET, old_value)

    def update(self, data=None, **kwargs):
        """
        Updates the data held within this object and triggers the
        appropriate events with :meth:`.changed`.
        """
        def trigger_changed(changed_object):
            try:
                items = changed_object.iteritems()
            except AttributeError:  # pragma: no cover
                items = changed_object.items()

            for key, value in items:
                if key not in self:
                    self.changed(self.CREATED, key, value, NOTSET)

                elif self[key] != value:
                    self.changed(self.MODIFIED, key, value, self[key])

        if isinstance(data, dict):
            for key, value in data.items():
                data[key] = self._map_value(key, value)

            trigger_changed(data)

        elif data is not None:
            raise TypeError("Expected None or dict for `data`")

        elif data is None:
            data = {}

        if kwargs:
            for key, value in kwargs.items():
                kwargs[key] = self._map_value(key, value)

            trigger_changed(kwargs)

        super(LoggingConfiguration, self).update(data, **kwargs)

    def changed(self, change_type, key, new_value=NOTSET, old_value=NOTSET):
        """
        This method is run whenever one of the keys in this object
        changes.
        """
        assert new_value is not NOTSET if change_type != self.DELETED else True
        assert old_value is NOTSET if change_type == self.CREATED else True

        if change_type == self.MODIFIED:
            logger.info("modified %r = %r", key, new_value)

        elif change_type == self.CREATED:
            logger.info("set %r = %r", key, new_value)

        elif change_type == self.DELETED:
            logger.warning("deleted %r", key)

        else:
            raise NotImplementedError(
                "Don't know how to handle change_type %r" % change_type)

    def master_contacted(self, update=True, announcement=False):
        """
        Simple method that will update the ``last_master_contact`` and then
        return the result.

        :param bool update:
            Setting this value to False will just return the current value
            instead of updating the value too.
        """
        if not update and "last_master_contact" not in self:
            return None

        if announcement:
            self["last_announce"] = datetime.utcnow()

        if update:
            self["last_master_contact"] = datetime.utcnow()

        return self["last_master_contact"]


class ConfigurationWithCallbacks(LoggingConfiguration):
    """
    Subclass of :class:`.LoggingDictionary` that provides the ability to
    run a function when a value is changed.
    """
    callbacks = {}

    @classmethod
    def register_callback(cls, key, callback, append=False):
        """
        Register a function as a callback for ``key``.  When ``key``
        is set the given ``callback`` will be run by :meth:`.changed`

        :param string key:
            the key which when changed in any way will execute
            ``callback``

        :param callable callback:
            the function or method to register

        :param boolean append:
            by default attempting to register a callback which has
            already been registered will do nothing, setting this
            to ``True`` overrides this behavior.
        """
        assert callable(callback)
        callbacks = cls.callbacks.setdefault(key, [])

        if callback in callbacks and not append:
            logger.warning(
                "%r is already a registered callback for %r", callback, key)
            return

        callbacks.append(callback)
        logger.debug("Registered callback %r for %r", callback, key)

    @classmethod
    def deregister_callback(cls, key, callback):
        """
        Removes any callback(s) that are registered with the provided ``key``
        """
        results = cls.callbacks.pop(key, None)
        if results is None:  # pragma: no cover
            logger.warning(
                "%r is not a registered callback for %r", callback, key)

    def clear(self, callbacks=False):
        """
        Performs the same operations as :meth:`dict.clear` except
        this method can also clear any registered callbacks if
        requested.
        """
        super(ConfigurationWithCallbacks, self).clear()
        if callbacks:
            self.callbacks.clear()

    def changed(self, change_type, key, new_value=NOTSET, old_value=NOTSET):
        """
        This method is called internally whenever a given ``key``
        changes which in turn will pass off the change to any
        registered callback(s).
        """
        super(ConfigurationWithCallbacks, self).changed(
            change_type, key, new_value=new_value, old_value=old_value)

        if key in self.callbacks:
            for callback in self.callbacks[key]:
                callback(change_type, key, new_value, old_value)
                logger.debug(
                    "Key %r was %r, calling callback %s",
                    key, change_type, callback)

# Prevent a call to reload() from dumping the config object
try:
    config
except NameError:
    config = ConfigurationWithCallbacks()


# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
# Copyright 2014 Ambient Entertainment GmbH & Co. KG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from httplib import (
        ACCEPTED, BAD_REQUEST, CONFLICT, SERVICE_UNAVAILABLE, OK)
except ImportError:  # pragma: no cover
    from http.client import (
        ACCEPTED, BAD_REQUEST, CONFLICT, SERVICE_UNAVAILABLE, OK)
from traceback import format_exception
from functools import partial

from twisted.web.server import NOT_DONE_YET
from twisted.internet import reactor
from voluptuous import Schema, Required

from pyfarm.core.enums import WorkState
from pyfarm.agent.config import config
from pyfarm.agent.http.core.client import post, http_retry_delay
from pyfarm.agent.http.api.base import APIResource
from pyfarm.agent.logger import getLogger
from pyfarm.agent.utility import request_from_master, uuid
from pyfarm.agent.sysinfo.memory import free_ram
from pyfarm.agent.utility import JOBTYPE_SCHEMA, TASKS_SCHEMA, JOB_SCHEMA
from pyfarm.jobtypes.core.jobtype import JobType

logger = getLogger("agent.http.assign")


class Assign(APIResource):
    isLeaf = False  # this is not really a collection of things

    # Schemas used for validating the request before
    # the target function will handle it.  These make
    # assertions about what kind of input data is required
    # or not based on the agent's internal code.
    SCHEMAS = {
        "POST": Schema({
            Required("job"): JOB_SCHEMA,
            Required("jobtype"): JOBTYPE_SCHEMA,
            Required("tasks"): TASKS_SCHEMA})}

    def __init__(self, agent):
        self.agent = agent

    def post(self, **kwargs):
        request = kwargs["request"]
        request_data = kwargs["data"]

        if request_from_master(request):
            config.master_contacted()

        # First, get the resources we have *right now*.  In some cases
        # this means using the functions in pyfarm.core.sysinfo because
        # entries in `config` could be slightly out of sync with the system.
        memory_free = free_ram()
        cpus = config["agent_cpus"]
        requires_ram = request_data["job"].get("ram")
        requires_cpus = request_data["job"].get("cpus")

        if self.agent.shutting_down:
            logger.error("Rejecting assignment because the agent is in the "
                         "process of shutting down.")
            request.setResponseCode(SERVICE_UNAVAILABLE)
            request.write(
                {"error": "Agent cannot accept assignments because it is "
                          "shutting down"})
            request.finish()
            return NOT_DONE_YET

        if "restart_requested" in config \
                and config["restart_requested"] is True:
            logger.error("Rejecting assignment because of scheduled restart.")
            request.setResponseCode(SERVICE_UNAVAILABLE)
            request.write(
                {"error": "Agent cannot accept assignments because of a "
                          "pending restart"})
            request.finish()
            return NOT_DONE_YET

        elif "agent-id" not in config:
            logger.error(
                "Agent has not yet connected to the master or `agent-id` "
                "has not been set yet.")
            request.setResponseCode(SERVICE_UNAVAILABLE)
            request.write(
                {"error": "agent-id has not been set in the config"})
            request.finish()
            return NOT_DONE_YET

        # Do we have enough ram?
        elif requires_ram is not None and requires_ram > memory_free:
            logger.error(
                "Task %s requires %sMB of ram, this agent has %sMB free.  "
                "Rejecting Task %s.",
                request_data["job"]["id"], requires_ram, memory_free,
                request_data["job"]["id"])
            request.setResponseCode(BAD_REQUEST)
            request.write(
                {"error": "Not enough ram",
                 "agent_ram": memory_free,
                 "requires_ram": requires_ram})
            request.finish()

            # touch the config
            config["free_ram"] = memory_free
            return NOT_DONE_YET

        # Do we have enough cpus (count wise)?
        elif requires_cpus is not None and requires_cpus > cpus:
            logger.error(
                "Task %s requires %s CPUs, this agent has %s CPUs.  "
                "Rejecting Task %s.",
                request_data["job"]["id"], requires_cpus, cpus,
                request_data["job"]["id"])
            request.setResponseCode(BAD_REQUEST)
            request.write(
                {"error": "Not enough cpus",
                 "agent_cpus": cpus,
                 "requires_cpus": requires_cpus})
            request.finish()
            return NOT_DONE_YET

        # Check for double assignments
        try:
            current_assignments = config["current_assignments"].itervalues
        except AttributeError:  # pragma: no cover
            current_assignments = config["current_assignments"].values

        existing_task_ids = set()
        new_task_ids = set(task["id"] for task in request_data["tasks"])

        for assignment in current_assignments():
            for task in assignment["tasks"]:
                existing_task_ids.add(task["id"])

        if existing_task_ids & new_task_ids:
            request.setResponseCode(CONFLICT)
            request.write(
                {"error": "Double assignment of tasks",
                 "duplicate_tasks": list(existing_task_ids & new_task_ids)})
            request.finish()
            return NOT_DONE_YET

        assignment_uuid = uuid()
        request_data.update(id=assignment_uuid)
        config["current_assignments"][assignment_uuid] = request_data

        # In all other cases we have some work to do inside of
        # deferreds so we just have to respond
        # TODO Mark this agent as running on the master
        request.setResponseCode(ACCEPTED)
        request.write({"id": assignment_uuid})
        request.finish()
        logger.info("Accepted assignment %s: %r", assignment_uuid, request_data)

        def assignment_failed(result, assign_id):
            logger.error(
                "Assignment %s failed, removing.", assign_id)
            logger.error(result.getTraceback())
            assignment = config["current_assignments"].pop(assign_id)
            if "jobtype" in assignment:
                jobtype_id = assignment["jobtype"].pop("id", None)
                if jobtype_id:
                    config["jobtypes"].pop(jobtype_id, None)

        def assignment_started(_, assign_id):
            logger.debug("Assignment %s has started", assign_id)

        def assignment_stopped(_, assign_id):
            logger.debug("Assignment %s has stopped", assign_id)
            assignment = config["current_assignments"].pop(assign_id)
            if "jobtype" in assignment:
                jobtype_id = assignment["jobtype"].pop("id", None)
                if jobtype_id:
                    config["jobtypes"].pop(jobtype_id, None)

        def restart_if_necessary(_):  # pragma: no cover
            if "restart_requested" in config and config["restart_requested"]:
                config["agent"].stop()

        def load_jobtype_failed(result, assign_id):
            logger.error(
                "Loading jobtype for assignment %s failed, removing.", assign_id)
            traceback = result.getTraceback()
            logger.debug("Got traceback")
            logger.error(traceback)
            assignment = config["current_assignments"].pop(assign_id)

            # Mark all tasks as failed on master and set an error message
            logger.debug("Marking tasks in assignment as failed")
            def post_update(post_url, post_data, task, delay=0):
                post_func = partial(post, post_url, data=post_data,
                    callback=lambda x: result_callback(
                        post_url, post_data, task, x),
                    errback=lambda x: error_callback(
                        post_url, post_data, task, x))
                reactor.callLater(delay, post_func)

            def result_callback(cburl, cbdata, task, response):
                if 500 <= response.code < 600:
                    logger.error(
                        "Error while marking task %s as failed on master, "
                        "retrying", task["id"])
                    post_update(cburl, cbdata, task, delay=http_retry_delay())

                elif response.code != OK:
                    logger.error(
                        "Could not mark task %s as failed, server response "
                        "code was %s", task["id"], response.code)

                else:
                    logger.info(
                        "Marked task %s as failed on master", task["id"])

            def error_callback(cburl, cbdata, task, failure_reason):
                logger.error(
                    "Error while marking task %s as failed, retrying",
                    task["id"], failure_reason)
                post_update(cburl, cbdata, task, delay=http_retry_delay())

            for task in assignment["tasks"]:
                url = "%s/jobs/%s/tasks/%s" % (
                    config["master_api"], assignment["job"]["id"], task["id"])
                data = {
                    "state": WorkState.FAILED,
                    "last_error": traceback}
                post_update(url, data, task)

            # If the loading was partially successful for some reason, there
            # might already be an entry for this jobtype in the config.
            # Remove it if it exists.
            if "jobtype" in assignment:
                jobtype_id = assignment["jobtype"].pop("id", None)
                if jobtype_id:
                    config["jobtypes"].pop(jobtype_id, None)

        def loaded_jobtype(jobtype_class, assign_id):
            # TODO: report error to master
            if hasattr(jobtype_class, "getTraceback"):
                logger.error(jobtype_class.getTraceback())
                return

            # TODO: add call to prepare_for_job
            # TODO: add call to spawn_persistent_process

            # Instance the job type and pass in the assignment data.
            instance = jobtype_class(request_data)

            if not isinstance(instance, JobType):
                raise TypeError(
                    "Expected a subclass of "
                    "pyfarm.jobtypes.core.jobtype.JobType")

            # TODO: add callback to cleanup_after_job
            # TODO: add callback to stop persistent process
            try:
                started_deferred, stopped_deferred = instance._start()
                started_deferred.addCallback(assignment_started, assign_id)
                started_deferred.addErrback(assignment_failed, assign_id)
                stopped_deferred.addCallback(assignment_stopped, assign_id)
                stopped_deferred.addErrback(assignment_failed, assign_id)
                stopped_deferred.addBoth(restart_if_necessary)
                stopped_deferred.addBoth(
                    lambda *args: instance._remove_tempdirs())
            except Exception as e:
                logger.error("Error on starting jobtype, stopping it now.  "
                             "Error was: %s", e)
                instance.stop()
                assignment = config["current_assignments"].pop(assign_id)
                if "jobtype" in assignment:
                    jobtype_id = assignment["jobtype"].pop("id", None)
                    if jobtype_id:
                        config["jobtypes"].pop(jobtype_id, None)

        # Load the job type then pass the class along to the
        # callback.  No errback here because all the errors
        # are handled internally in this case.
        jobtype_loader = JobType.load(request_data)
        jobtype_loader.addCallback(loaded_jobtype, assignment_uuid)
        jobtype_loader.addErrback(load_jobtype_failed, assignment_uuid)

        return NOT_DONE_YET



# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
User
----

Returns information about the current user such as the user name, admin
access, or other related information.
"""

import os
import ctypes


# Linux/mac
try:
    import pwd
except ImportError:  # pragma: no cover
    pwd = NotImplemented

# Windows
try:  # pragma: no cover
    import win32api
    from win32com.shell import shell
except ImportError:  # pragma: no cover
    win32api = NotImplemented

# Shouldn't generally need this, but just in case someone
# has a broken setup we should try
try:
    import getpass
except ImportError:  # pragma: no cover
    getpass = NotImplemented

try:
    from os import getuid
except ImportError:  # pragma: no cover
    getuid = NotImplemented


from pyfarm.agent.sysinfo.system import operating_system
from pyfarm.core.enums import WINDOWS


def username():
    """
    Returns the current user name using the most native api
    we can import. On Linux for example this will use the :mod:`pwd`
    module but on Windows we try to use :mod:`win32api`.
    """
    if pwd is not NotImplemented:
        return pwd.getpwuid(os.getuid())[0]
    elif win32api is not NotImplemented:  # pragma: no cover
        return win32api.GetUserName()
    elif getpass is not NotImplemented:  # pragma: no cover
        return getpass.getuser()
    else:
        raise NotImplementedError("neither `getpass` or `pwd` were imported")


def is_administrator():
    """
    Return True if the current user is root (Linux) or running as an
    Administrator (Windows).
    """
    if getuid is not NotImplemented:
        return getuid() == 0
    elif win32api is not NotImplemented:  # pragma: no cover
        return shell.IsUserAnAdmin()
    elif win32api is NotImplemented and WINDOWS:  # pragma: no cover
        return ctypes.windll.shell32.IsUserAnAdmin() != 0
    else:
        raise NotImplementedError(
            "`is_administrator` is not implemented for %s" % operating_system())



# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
HTTP Server
-----------

HTTP server responsible for serving requests that
control or query the running agent.  This file produces
a service that the  :class:`pyfarm.agent.manager.service.ManagerServiceMaker`
class can consume on start.
"""

import re

try:
    from httplib import FORBIDDEN
except ImportError:  # pragma: no cover
    from http.client import FORBIDDEN

from os.path import exists

from twisted.web.server import Site as _Site, Request as _Request
from twisted.web.static import File
from twisted.web.error import Error

from pyfarm.core.enums import STRING_TYPES
from pyfarm.agent.utility import dumps


class RewriteRequest(_Request):
    """
    A custom implementation of :class:`._Request` that will allow us
    to modify an incoming request before it reaches the HTTP server..
    """
    REPLACE_REPEATED_DELIMITER = re.compile("/{2,}")

    def requestReceived(self, command, path, version):
        """
        Override the built in :meth:`._Request.requestReceived` so we
        can rewrite portions of the request, such as the url, before it's
        passed along to the internal server.
        """
        # before we give the path to Twisted, replace any
        # repeated `/`s with `/`
        if "//" in path:
            path = self.REPLACE_REPEATED_DELIMITER.sub("/", path)
        _Request.requestReceived(self, command, path, version)

    def write(self, data):
        """
        Override the built in :meth:`._Request.write` so that any data
        that's not a string will be dumped to json using :func:`.dumps`
        """
        if not isinstance(data, STRING_TYPES):
            data = dumps(data)

        _Request.write(self, data)


class Site(_Site):
    """
    Site object similar to Twisted's except it also carries along
    some of the internal agent data.
    """
    displayTracebacks = True
    requestFactory = RewriteRequest


class StaticPath(File):
    """
    More secure version of :class:`.File` that does not list
    directories.  In addition this will also sending along a
    response header asking clients to cache to data.
    """
    EXPIRES = 604800  # 7 days
    ALLOW_DIRECTORY_LISTING = False

    def __init__(self, *args, **kwargs):
        File.__init__(self, *args, **kwargs)

        if not exists(self.path):
            raise OSError("%s does not exist" % self.path)

    def render(self, request):
        """Overrides :meth:`.File.render` and sets the expires header"""
        request.responseHeaders.addRawHeader(
            "Cache-Control", "max-age=%s" % self.EXPIRES)
        return File.render(self, request)

    def directoryListing(self):
        """Override which ensures directories cannot be listed"""
        if not self.ALLOW_DIRECTORY_LISTING:
            raise Error(FORBIDDEN, "directory listing is not allowed")
        return File.directoryListing(self)




# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Base
++++

Contains the base resources used for building up the root
of the agent's api.
"""

from pyfarm.agent.config import config
from pyfarm.agent.http.core.resource import Resource
from pyfarm.agent.utility import dumps, request_from_master


class APIResource(Resource):
    """Base class for all api resources"""
    isLeaf = True
    CONTENT_TYPES = set(["application/json"])


class APIRoot(APIResource):
    isLeaf = False


class Versions(APIResource):
    """
    Returns a list of api versions which this agent will support

    .. http:get:: /api/v1/versions/ HTTP/1.1

        **Request**

        .. sourcecode:: http

            GET /api/v1/versions/HTTP/1.1
            Accept: application/json

        **Response**

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "versions": [1]
            }

    """
    isLeaf = True

    def get(self, **kwargs):
        request = kwargs.get("request")

        if request is not None and request_from_master(request):
            config.master_contacted()

        return dumps(versions=[1])



# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Network
-------

Returns information about the network including ip address, dns, data
sent/received, and some error information.

:const IP_PRIVATE:
    set of private class A, B, and C network ranges

    .. seealso::
        :rfc:`1918`

:const IP_NONNETWORK:
    set of non-network address ranges including all of the
    above constants except the :const:`IP_PRIVATE`
"""

import socket

import netifaces
from netaddr import IPSet, IPNetwork, IPAddress

from pyfarm.agent.logger import getLogger

logger = getLogger("agent.dns")

IP_PRIVATE = IPSet([
    IPNetwork("10.0.0.0/8"),
    IPNetwork("172.16.0.0/12"),
    IPNetwork("192.168.0.0/16")])
IP_NONNETWORK = IPSet([
    IPNetwork("0.0.0.0/8"),         # special use
    IPNetwork("169.254.0.0/16"),    # link local
    IPNetwork("127.0.0.0/8"),       # loopback
    IPNetwork("224.0.0.0/4"),       # multicast
    IPNetwork("255.255.255.255")])  # broadcast


def mac_addresses(long_addresses=False, as_integers=False):
    """
    Returns a tuple of all mac addresses on the system.

    :param bool standard_length_only:
        Some adapters will specify a mac address which is longer
        than the standard value of six pairs.  Setting this value
        to False will allow these to be displayed.

    :param bool as_integers:
        When ``True`` convert all mac addresses to integers.
    """
    results = set()
    for ifaces in map(netifaces.ifaddresses, netifaces.interfaces()):
        for entry in ifaces.get(netifaces.AF_LINK, []):
            mac = entry.get("addr", "")

            if all([mac, not long_addresses, len(mac) == 17]) \
                    or not all([not long_addresses, mac]):
                if as_integers:
                    mac = 0 if not mac else int("0x" + mac.replace(":", ""), 0)
                results.add(mac)

    return tuple(results)


def hostname(trust_name_from_ips=True):
    """
    Returns the hostname which the agent should send to the master.

    :param bool trust_resolved_name:
        If True and all addresses provided by :func:`addresses` resolve
        to a single hostname then just return that name as it's the most
        likely hostname to be accessible by the rest of the network.
    """
    logger.debug("Attempting to discover the hostname")

    # For every address retrieve the hostname we can resolve it
    # to.  We'll use this set later to compare again what the system
    # is telling is the hostname should be.
    reverse_hostnames = set()
    ip_addresses = addresses()
    for address in ip_addresses:
        try:
            dns_name, aliases, dns_addresses = socket.gethostbyaddr(address)

        except socket.herror:  # pragma: no cover
            logger.warning(
                "Could not resolve %s to a hostname using DNS.", address)
        else:
            if address in dns_addresses:
                reverse_hostnames.add(dns_name)
                logger.debug(
                    "Lookup of %s resolved to %s", address, dns_name)

    # If all the addresses we know about map to a single
    # hostname then just return that instead of continuing
    # on.  We do this because the DNS system should know better
    # than the local system will what the fully qualified
    # hostname would be.  This is especially true on some
    # platforms where the local DNS implementation seems to
    # produce the wrong information when resolving the local
    # hostname.
    if len(reverse_hostnames) == 1 and trust_name_from_ips:
        return reverse_hostnames.pop()

    if not reverse_hostnames:
        logger.warning(
            "DNS failed to resolve %s to hostnames", ip_addresses)

    local_hostname = socket.gethostname()
    local_fqdn_query = socket.getfqdn()

    if local_fqdn_query in reverse_hostnames:  # pragma: no cover
        name = local_fqdn_query

    elif local_hostname in reverse_hostnames:  # pragma: no cover
        name = local_hostname

    else:
        # If neither local_hostname or local_fqdn_query seemed
        # to map to a known hostname from an IP address then
        # get the FQDN of the locally provided hostname as
        # a fallback.
        name = socket.getfqdn(local_hostname)

    if name.startswith("localhost"):
        logger.warning("Hostname resolved to or contains 'localhost'")

    if name.endswith(".local"):
        logger.warning(
            "Operating system '.local' to hostname.  In some cases, most "
            "often on OS X, this may cause unexpected problems reaching this "
            "host by name on the network.  Manually setting the hostname "
            "with --hostname may be advisable.")

    return name


def addresses(private_only=True):
    """Returns a tuple of all non-local ip addresses."""
    results = set()

    for interface in netifaces.interfaces():
        addrinfo = netifaces.ifaddresses(interface)
        for address in addrinfo.get(socket.AF_INET, []):
            addr = address.get("addr")

            if addr is not None:
                # Make sure that what we're getting out of
                # netifaces is something we can use.
                try:
                    ip = IPAddress(addr)
                except ValueError:  # pragma: no cover
                    logger.error(
                        "Could not convert %s to a valid IP object" % addr)
                else:
                    if ip in IP_PRIVATE or not private_only:
                        results.add(addr)

    if not addresses:  # pragma: no cover
        logger.error("No addresses could be found")

    return tuple(results)


def interfaces():
    """Returns the names of all valid network interface names"""
    results = set()

    for name in netifaces.interfaces():
        # only add network interfaces which have IPv4
        addresses = netifaces.ifaddresses(name)

        if socket.AF_INET not in addresses:  # pragma: no cover
            continue

        if any(addr.get("addr") for addr in addresses[socket.AF_INET]):
            results.add(name)

    if not results:  # pragma: no cover
        logger.warning("Failed to find any interfaces")

    return tuple(results)




# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
# Copyright 2014 Ambient Entertainment GmbH & Co. KG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from json import dumps

try:
    from httplib import (
        BAD_REQUEST, NO_CONTENT, INTERNAL_SERVER_ERROR, NOT_FOUND, ACCEPTED)
except ImportError:  # pragma: no cover
    from http.client import (
        BAD_REQUEST, NO_CONTENT, INTERNAL_SERVER_ERROR, NOT_FOUND, ACCEPTED)

from twisted.web.server import NOT_DONE_YET

from pyfarm.agent.config import config
from pyfarm.agent.http.api.base import APIResource
from pyfarm.agent.utility import request_from_master
from pyfarm.agent.logger import getLogger

logger = getLogger("agent.http.api.tasks")


class Tasks(APIResource):
    def get(self, **kwargs):
        request = kwargs["request"]

        if request_from_master(request):
            config.master_contacted()

        try:
            current_assignments = config["current_assignments"].itervalues
        except AttributeError:  # pragma: no cover
            current_assignments = config["current_assignments"].values

        tasks = []
        for assignment in current_assignments():
            tasks += assignment["tasks"]

        return dumps(tasks)

    def delete(self, **kwargs):
        """
        HTTP endpoint for stopping and deleting an individual task from this
        agent.
        ... warning::
        If the specified task is part of a multi-task assignment, all
        tasks in this assignment will be stopped, not just the specified one.

        This will try to asynchronously stop the assignment by killing all its
        child processes. If that isn't successful, this will have no effect.
        """
        request = kwargs["request"]

        # Postpath must be exactly one element, and that needs to be an integer
        if len(request.postpath) != 1:
            request.setResponseCode(BAD_REQUEST)
            request.write({"error": "Did not specify a task id"})
            return NOT_DONE_YET

        try:
            task_id = int(request.postpath[0])
        except ValueError:
            request.setResponseCode(BAD_REQUEST)
            request.write({"error": "Task id was not an integer"})
            return NOT_DONE_YET

        try:
            current_assignments = config["current_assignments"].itervalues
        except AttributeError:  # pragma: no cover
            current_assignments = config["current_assignments"].values

        assignment = None
        for item in current_assignments():
            for task in item["tasks"]:
                if task["id"] == task_id:
                    assignment = item

        if not assignment:
            logger.info("Cannot cancel task %s: not found", task_id)
            request.setResponseCode(NO_CONTENT)
            request.finish()
            return NOT_DONE_YET

        if "jobtype" in assignment and "id" in assignment["jobtype"]:
            jobtype = config["jobtypes"][assignment["jobtype"]["id"]]
            logger.info("Stopping assignment %s", assignment["id"])
            jobtype.stop()
        else:
            logger.error("Tried stopping assigment %s, but found no jobtype "
                         "instance", assignment.id)
            request.setResponseCode(INTERNAL_SERVER_ERROR)
            request.write({"error": "Assignment found, but no jobtype instance "
                                    "exists."})
            request.finish()
            return NOT_DONE_YET

        request.setResponseCode(ACCEPTED)
        request.finish()

        return NOT_DONE_YET



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Development
-----------

Contains entry points which constructs the command line tools
used for development such as ``pyfarm-dev-fakerender`` and
``pyfarm-dev-fakework``.
"""

import argparse
import ctypes
import hashlib
import os
import sys
import time
from collections import namedtuple
from functools import partial
from random import choice, randint, random
from textwrap import dedent

import psutil
import requests

from pyfarm.core.enums import NUMERIC_TYPES
from pyfarm.core.utility import convert
from pyfarm.agent.config import config
from pyfarm.agent.entrypoints.parser import integer, number
from pyfarm.agent.logger import getLogger
from pyfarm.agent.utility import dumps


logger = getLogger("agent.cmd")

config["start"] = time.time()


def fake_render():
    process = psutil.Process()
    memory_usage = lambda: convert.bytetomb(process.get_memory_info().rss)
    memory_used_at_start = memory_usage()
    FakeInstance = namedtuple("FakeInstance", ("parser", "args"))

    logger.info("sys.argv: %r", sys.argv)

    # build parser
    parser = argparse.ArgumentParser(
        description="Very basic command line tool which vaguely simulates a "
                    "render.")
    getnum = partial(
        number, instance=FakeInstance(parser=parser, args=None),
        types=NUMERIC_TYPES,
        allow_inf=False, min_=0)
    parser.add_argument(
        "--ram", type=getnum, default=25,
        help="How much ram in megabytes the fake command should consume")
    parser.add_argument(
        "--duration", type=getnum, default=5,
        help="How many seconds it should take to run this command")
    parser.add_argument(
        "--return-code", type=getnum, action="append", default=[0],
        help="The return code to return, declaring this flag multiple times "
             "will result in a random return code.  [default: %(default)s]")
    parser.add_argument(
        "--duration-jitter", type=getnum, default=5,
        help="Randomly add or subtract this amount to the total duration")
    parser.add_argument(
        "--ram-jitter", type=getnum, default=None,
        help="Randomly add or subtract this amount to the ram")
    parser.add_argument(
        "-s", "--start", type=getnum, required=True,
        help="The start frame.  If no other flags are provided this will also "
             "be the end frame.")
    parser.add_argument(
        "-e", "--end", type=getnum, help="The end frame")
    parser.add_argument(
        "-b", "--by", type=getnum, help="The by frame", default=1)
    parser.add_argument(
        "--spew", default=False, action="store_true",
        help="Spews lots of random output to stdout which is generally "
             "a decent stress test for log processing issues.  Do note however "
             "that this will disable the code which is consuming extra CPU "
             "cycles.  Also, use this option with care as it can generate "
             "several gigabytes of data per frame.")
    parser.add_argument(
        "--segfault", action="store_true",
        help="If provided then there's a 25%% chance of causing a segmentation "
             "fault.")
    args = parser.parse_args()

    if args.end is None:
        args.end = args.start

    if args.ram_jitter is None:
        args.ram_jitter = int(args.ram / 2)

    assert args.end >= args.start and args.by >= 1

    random_output = None
    if args.spew:
        random_output = list(os.urandom(1024).encode("hex") for _ in xrange(15))

    errors = 0
    if isinstance(args.start, float):
        logger.warning(
            "Truncating `start` to an integer (float not yet supported)")
        args.start = int(args.start)

    if isinstance(args.end, float):
        logger.warning(
            "Truncating `end` to an integer (float not yet supported)")
        args.end = int(args.end)

    if isinstance(args.by, float):
        logger.warning(
            "Truncating `by` to an integer (float not yet supported)")
        args.by = int(args.by)

    for frame in xrange(args.start, args.end + 1, args.by):
        duration = args.duration + randint(
            -args.duration_jitter, args.duration_jitter)
        ram_usage = max(
            0, args.ram + randint(-args.ram_jitter, args.ram_jitter))
        logger.info("Starting frame %04d", frame)

        # Warn if we're already using more memory
        if ram_usage < memory_used_at_start:
            logger.warning(
                "Memory usage starting up is higher than the value provided, "
                "defaulting --ram to %s", memory_used_at_start)

        # Consume the requested memory (or close to)
        # TODO: this is unrealistic, majority of renders don't eat ram like this
        memory_to_consume = int(ram_usage - memory_usage())
        big_string = " " * 1048576  # ~ 1MB of memory usage
        if memory_to_consume > 0:
            start = time.time()
            logger.debug(
                "Consuming %s megabytes of memory", memory_to_consume)
            try:
                big_string += big_string * memory_to_consume

            except MemoryError:
                logger.error("Cannot render, not enough memory")
                errors += 1
                continue

            logger.debug(
                "Finished consumption of memory in %s seconds.  Off from "
                "target memory usage by %sMB.",
                time.time() - start, memory_usage() - ram_usage)

        # Decently guaranteed to cause a segmentation fault.  Originally from:
        #   https://wiki.python.org/moin/CrashingPython#ctypes
        if args.segfault and random() > .25:
            i = ctypes.c_char('a')
            j = ctypes.pointer(i)
            c = 0
            while True:
                j[c] = 'a'
                c += 1
            j

        # Continually hash a part of big_string to consume
        # cpu cycles
        end_time = time.time() + duration
        last_percent = None
        while time.time() < end_time:
            if args.spew:
                print >> sys.stdout, choice(random_output)
            else:
                hashlib.sha1(big_string[:4096])  # consume CPU cycles

            progress = (1 - (end_time - time.time()) / duration) * 100
            percent, _ = divmod(progress, 5)
            if percent != last_percent:
                last_percent = percent
                logger.info("Progress %03d%%", progress)

        if last_percent is None:
            logger.info("Progress 100%%")

        logger.info("Finished frame %04d in %s seconds", frame, duration)

    if errors:
        logger.error("Render finished with errors")
        sys.exit(1)
    else:
        return_code = choice(args.return_code)
        logger.info("exit %s", return_code)


def fake_work():
    parser = argparse.ArgumentParser(
        description="Quick and dirty script to create a job type, a job, and "
                    "some tasks which are then posted directly to the "
                    "agent.  The primary purpose of this script is to test "
                    "the internal of the job types")
    FakeInstance = namedtuple("FakeInstance", ("parser", "args"))
    getint = partial(
        integer, instance=FakeInstance(parser=parser, args=None),
        allow_inf=False, min_=0)
    parser.add_argument(
        "--master-api", default="http://127.0.0.1/api/v1",
        help="The url to the master's api [default: %(default)s]")
    parser.add_argument(
        "--agent-api", default="http://127.0.0.1:50000/api/v1",
        help="The url to the agent's api [default: %(default)s]")
    parser.add_argument(
        "--jobtype", default="FakeRender",
        help="The job type to use [default: %(default)s]")
    parser.add_argument(
        "--job", type=getint,
        help="If provided then this will be the job we pull tasks from "
             "and assign to the agent.  Please note we'll only be pulling "
             "tasks that aren't running or assigned.")
    args = parser.parse_args()
    logger.info("Master args.master_api: %s", args.master_api)
    logger.info("Agent args.master_api: %s", args.agent_api)
    assert not args.agent_api.endswith("/")
    assert not args.master_api.endswith("/")

    # session to make requests with
    session = requests.Session()
    session.headers.update({"content-type": "application/json"})

    existing_jobtype = session.get(
        args.master_api + "/jobtypes/%s" % args.jobtype)

    # Create a FakeRender job type if one does not exist
    if not existing_jobtype.ok:
        sourcecode = dedent("""
        from pyfarm.jobtypes.examples import %s as _%s
        class %s(_%s):
            pass""" % (args.jobtype, args.jobtype, args.jobtype, args.jobtype))
        response = session.post(
            args.master_api + "/jobtypes/",
            data=dumps({
                "name": args.jobtype,
                "classname": args.jobtype,
                "code": sourcecode,
                "max_batch": 1}))
        assert response.ok, response.json()
        jobtype_data = response.json()
        logger.info(
            "Created job type %r, id %r", args.jobtype, jobtype_data["id"])

    else:
        jobtype_data = existing_jobtype.json()
        logger.info(
            "Job type %r already exists, id %r",
            args.jobtype, jobtype_data["id"])

    jobtype_version = jobtype_data["version"]

    if args.job is None:
        job = session.post(
            args.master_api + "/jobs/",
            data=dumps({
                "start": 1,
                "end": 1,
                "title": "Fake Job - %s" % int(time.time()),
                "jobtype": args.jobtype}))
        assert job.ok, job.json()
        job = job.json()
        logger.info("Job %r created", job["id"])
    else:
        job = session.get(args.master_api + "/jobs/%s" % args.job)
        if not job.ok:
            logger.error("No such job with id %r", args.job)
            return
        else:
            job = job.json()
            logger.info("Job %r exists", job["id"])

    tasks = session.get(args.master_api + "/jobs/%s/tasks/" % job["id"])
    assert tasks.ok

    job_tasks = []
    for task in tasks.json():
        if task["state"] not in ("queued", "failed"):
            logger.info(
                "Can't use task %s, it's state is not 'queued' or 'failed'",
                task["id"])
            continue

        if task["agent_id"] is not None:
            logger.info(
                "Can't use task %s, it already has an agent assigned",
                task["id"])

        job_tasks.append({"id": task["id"], "frame": task["frame"]})

    if not job_tasks:
        logger.error("Could not find any tasks to send for job %s", job["id"])
        return

    logger.info(
        "Found %s tasks from job %s to assign to %r",
        len(job_tasks), job["id"], args.agent_api)

    assignment_data = {
        "job": {
            "id": job["id"],
            "by": job["by"],
            "ram": job["ram"],
            "ram_warning": job["ram_warning"],
            "ram_max": job["ram_max"],
            "cpus": job["cpus"],
            "batch": job["batch"],
            "user": job["user"],
            "data": job["data"],
            "environ": job["environ"],
            "title": job["title"]},
        "jobtype": {
            "name": args.jobtype,
            "version": jobtype_version},
        "tasks": job_tasks}

    # Drop any keys which don't have values since this
    # would break the schema validation in the agent.
    for key in list(assignment_data["job"]):
        if assignment_data["job"][key] is None:
            del assignment_data["job"][key]

    response = session.post(
        args.agent_api + "/assign",
        data=dumps(assignment_data))
    assert response.ok, response.json()
    logger.info("Tasks posted to agent")




# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Process
--------

Module responsible for connecting a Twisted process object and
a job type.  Additionally this module contains other classes which
are useful in starting or managing a process.
"""

import os

from psutil import Process, NoSuchProcess
from twisted.internet import reactor
from twisted.internet.protocol import ProcessProtocol as _ProcessProtocol

from pyfarm.agent.logger import getLogger
from pyfarm.agent.utility import uuid
from pyfarm.jobtypes.core.log import STDOUT, STDERR

logger = getLogger("jobtypes.process")


class ReplaceEnvironment(object):
    """
    A context manager which will replace ``os.environ``'s, or dictionary of
    your choosing, for a short period of time.  After exiting the
    context manager the original environment will be restored.

    This is useful if you have something like a process that's using
    global environment and you want to ensure that global environment is
    always consistent.

    :param dict environment:
        If provided, use this as the environment dictionary instead
        of ``os.environ``
    """
    def __init__(self, frozen_environment, environment=None):
        if environment is None:
            environment = os.environ

        self.environment = environment
        self.original_environment = None
        self.frozen_environment = frozen_environment

    def __enter__(self):
        self.original_environment = self.environment.copy()
        self.environment.clear()
        self.environment.update(self.frozen_environment)
        return self

    def __exit__(self, *_):
        self.environment.clear()
        self.environment.update(self.original_environment)


class ProcessProtocol(_ProcessProtocol):
    """
    Subclass of :class:`.Protocol` which hooks into the various systems
    necessary to run and manage a process.  More specifically, this helps
    to act as plumbing between the process being run and the job type.
    """
    def __init__(self, jobtype):
        self.jobtype = jobtype

        # Internal only attributes, __uuid itself is a property so
        # it can't be accidentally modified later.
        self.__ended = False
        self.__uuid = uuid()

    def __repr__(self):
        return \
            "ProcessProtocol(uuid=%s, pid=%r)" % (self.uuid, self.pid)

    @property
    def uuid(self):
        return self.__uuid

    @property
    def pid(self):
        return self.transport.pid

    @property
    def process(self):
        """The underlying Twisted process object"""
        return self.transport

    @property
    def psutil_process(self):
        """Returns a :class:`psutil.Process` object for the running process"""
        # It's possible that the process could have
        # ended but not died yet.  So we have this
        # check just in case this property is called
        # during this state.
        if self.__ended:
            return None

        try:
            return Process(pid=self.pid)
        except NoSuchProcess:  # pragma: no cover
            return None

    def connectionMade(self):
        """
        Called when the process first starts and the file descriptors
        have opened.
        """
        self.jobtype._process_started(self)

    def processEnded(self, reason):
        """
        Called when the process has terminated and all file descriptors
        have been closed.  :meth:`processExited` is called, too, however we
        only want to notify the parent job type once the process has freed
        up the last bit of resources.
        """
        try:
            self.__ended = True
            self.jobtype._process_stopped(self, reason)
        except Exception as e:
            logger.error("Exception caught while running "
                         "jobtype._process_stopped: %s", e)

    def outReceived(self, data):
        """Called when the process emits on stdout"""
        try:
            self.jobtype._process_output(self, data, STDOUT)
        except Exception as e:
            logger.error("Caught an exception from _process_output: %s", e)

    def errReceived(self, data):
        """Called when the process emits on stderr"""
        try:
            self.jobtype._process_output(self, data, STDERR)
        except Exception as e:
            logger.error("Caught an exception from _process_output: %s", e)

    def kill(self):
        """Kills the underlying process, if running."""
        logger.info("Killing %s", self)
        try:
            process = self.psutil_process
            if not process:
                return
            children = process.children(recursive=True)
            process.kill()
        except Exception as e:  # pragma: no cover
            logger.warning("Cannot kill %s: %s.", self, e)

        def kill_children(children):
            for child in children:
                child.kill()
        if children:
            reactor.callLater(2, kill_children, children)

    def terminate(self):
        """Terminates the underlying process, if running."""
        logger.info("Terminating %s", self)
        try:
            process = self.psutil_process
            if not process:
                return
            children = process.children(recursive=True)
            process.terminate()
        except Exception as e:  # pragma: no cover
            logger.warning("Cannot terminate %s: %s.", self, e)

        def terminate_children(children):
            for child in children:
                child.terminate()
        if children:
            reactor.callLater(2, terminate_children, children)

    def interrupt(self):
        """Interrupts the underlying process, if running."""
        logger.info("Interrupt %s", self)
        try:
            self.process.signalProcess("INT")
        except Exception as e:  # pragma: no cover
            logger.warning("Cannot interrupt %s: %s.", self, e)

    def running(self):
        """Method to determine whether the child process is currently running"""
        return self.pid is not None



# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import with_statement

import sys
assert sys.version_info[0:2] >= (2, 6), "Python 2.6 or higher is required"

import os
from os import walk
from os.path import isfile, join
from setuptools import setup

install_requires = [
    "pyfarm.core",
    # "PyOpenSSL", "service_identity",  # required for full SSL support
    "netaddr", "twisted", "ntplib", "requests!=2.4.0", "treq",
    "voluptuous", "jinja2", "psutil>=2.1.0",
    "netifaces>=0.10.2"]

if "READTHEDOCS" in os.environ:
    install_requires += ["sphinxcontrib-httpdomain", "sphinx"]

# Windows is a little special because we have to have pywin32
# installed.  pyfarm.core uses it and certain components of
# other libraries use it too, such as twisted, so we check for
# it here.  Unfortunately, we can't use PyPi for this.
if sys.platform.startswith("win"):
    try:
        import win32api
    except ImportError:
        raise ImportError(
            "On Windows, you must manually install pywin32 before running "
            "pyfarm.core's setup.py.  This is required because there's not "
            "a package that we can pull down and reliably install from the "
            "Python package repository.  Please visit "
            "http://sourceforge.net/projects/pywin32/files/pywin32/ to "
            "download and install this package.")

if sys.version_info[0:2] == (2, 6):
    install_requires += ["importlib", "ordereddict", "argparse"]

if isfile("README.rst"):
    with open("README.rst", "r") as readme:
        long_description = readme.read()
else:
    long_description = ""


def get_package_data(parent, roots):
    output = []
    for top in roots:
        if top.startswith("/"):
            raise ValueError("get_package_data was given an absolute path or "
                             "the filesystem root to traverse, refusing.")
        for root, dirs, files in walk(top):
            for filename in files:
                output.append(join(root, filename).split(parent)[-1][1:])

    return output

agent_root = join("pyfarm", "agent")
agent_package_data_roots = (
    join(agent_root, "etc"),
    join(agent_root, "http", "static"),
    join(agent_root, "http", "templates"))

jobtype_root = join("pyfarm", "jobtypes")
jobtype_root_package_data_roots = (
    join(jobtype_root, "etc"), )

setup(
    name="pyfarm.agent",
    version="0.8.0",
    packages=[
        "pyfarm",
        "pyfarm.agent",
        "pyfarm.agent.entrypoints",
        "pyfarm.agent.http",
        "pyfarm.agent.http.api",
        "pyfarm.agent.http.core",
        "pyfarm.agent.logger",
        "pyfarm.agent.sysinfo",
        "pyfarm.jobtypes",
        "pyfarm.jobtypes.core"],
    data_files=[
        ("etc/pyfarm", [
            "pyfarm/jobtypes/etc/jobtypes.yml",
            "pyfarm/agent/etc/agent.yml"])],
    package_data={
        "pyfarm.agent": get_package_data(
            agent_root, agent_package_data_roots),
        "pyfarm.jobtypes": get_package_data(
            jobtype_root, jobtype_root_package_data_roots)},
    namespace_packages=["pyfarm"],
    entry_points={
        "console_scripts": [
            "pyfarm-agent = pyfarm.agent.entrypoints.main:agent",
            "pyfarm-supervisor = "
            "   pyfarm.agent.entrypoints.supervisor:supervisor",
            "pyfarm-dev-fakerender = "
            "   pyfarm.agent.entrypoints.development:fake_render",
            "pyfarm-dev-fakework = "
            "   pyfarm.agent.entrypoints.development:fake_work"]},
    include_package_data=True,
    install_requires=install_requires,
    url="https://github.com/pyfarm/pyfarm-agent",
    license="Apache v2.0",
    author="Oliver Palmer",
    author_email="development@pyfarm.net",
    description="Core module containing code to run PyFarm's agent.",
    long_description=long_description,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: No Input/Output (Daemon)",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Topic :: System :: Distributed Computing"])

if sys.platform.startswith("win"):
    print("WARNING:  Please be sure you've install the OpenSSL Library as "
          "some modules may break on Windows without it.")



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
API
---

This package contains the components that form the agent's api.  The objects
contained in this section act as an interface layer between an incoming
http request and the agent's internals.
"""


# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from pyfarm.jobtypes.core.jobtype import JobType, CommandData


class PythonHelloWorld(JobType):
    def get_command_data(self):
        return CommandData(
            sys.executable,
            "-c", "import time; time.sleep(3); print 'hello world'")



# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
# Copyright 2014 Ambient Entertainment GmbH & Co. KG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Manager Service
---------------

Sends and receives information from the master and performs systems level tasks
such as log reading, system information gathering, and management of processes.
"""

import atexit
import os
import time
from datetime import datetime, timedelta
from functools import partial
from os.path import join, isfile
from random import random

try:
    from httplib import (
        responses, OK, CREATED, NOT_FOUND, INTERNAL_SERVER_ERROR, BAD_REQUEST)
except ImportError:  # pragma: no cover
    from http.client import (
        responses, OK, CREATED, NOT_FOUND, INTERNAL_SERVER_ERROR, BAD_REQUEST)

from ntplib import NTPClient
from twisted.internet import reactor
from twisted.internet.defer import Deferred, DeferredList
from twisted.internet.error import ConnectionRefusedError

from pyfarm.core.enums import AgentState
from pyfarm.agent.config import config
from pyfarm.agent.http.api.assign import Assign
from pyfarm.agent.http.api.base import APIRoot, Versions
from pyfarm.agent.http.api.config import Config
from pyfarm.agent.http.api.state import Status, Stop
from pyfarm.agent.http.api.log import LogQuery
from pyfarm.agent.http.api.tasks import Tasks
from pyfarm.agent.http.api.tasklogs import TaskLogs
from pyfarm.agent.http.api.update import Update
from pyfarm.agent.http.core.client import post, http_retry_delay
from pyfarm.agent.http.core.resource import Resource
from pyfarm.agent.http.core.server import Site, StaticPath
from pyfarm.agent.http.system import Index, Configuration
from pyfarm.agent.logger import getLogger
from pyfarm.agent.tasks import ScheduledTaskManager
from pyfarm.agent.sysinfo import memory

svclog = getLogger("agent.service")
ntplog = getLogger("agent.service.ntp")


class Agent(object):
    """
    Main class associated with getting getting the internals of the
    agent's operations up and running including adding or updating
    itself with the master, starting the periodic task manager,
    and handling shutdown conditions.
    """
    def __init__(self):
        # so parts of this instance are accessible elsewhere
        assert "agent" not in config
        config["agent"] = self
        self.services = {}
        self.register_shutdown_events = False
        self.last_free_ram_post = time.time()
        self.shutting_down = False

        # reannounce_client is set when the agent id is
        # first set. reannounce_client_instance ensures
        # that once we start the announcement process we
        # won't try another until we're finished
        self.reannounce_client_request = None

        # Setup scheduled tasks
        self.scheduled_tasks = ScheduledTaskManager()
        self.scheduled_tasks.register(
            self.reannounce, config["agent_master_reannounce"])

    @classmethod
    def agent_api(cls):
        """
        Return the API url for this agent or None if `agent-id` has not
        been set
        """
        try:
            return cls.agents_endpoint() + str(config["agent-id"])
        except KeyError:
            svclog.error(
                "The `agent-id` configuration value has not been set yet")
            return None

    @classmethod
    def agents_endpoint(cls):
        """
        Returns the API endpoint for used for updating or creating
        agents on the master
        """
        return config["master_api"] + "/agents/"

    def should_reannounce(self):
        """Small method which acts as a trigger for :meth:`reannounce`"""
        if self.reannounce_client_request is not None:
            svclog.debug("Agent is already trying to announce itself.")
            return

        if self.shutting_down:
            return False

        contacted = config.master_contacted(update=False)
        remaining = (datetime.utcnow() - contacted).total_seconds()
        return remaining > config["agent_master_reannounce"]

    def reannounce(self):
        """
        Method which is used to periodically contact the master.  This
        method is generally called as part of a scheduled task.
        """
        if not self.should_reannounce():
            return

        svclog.debug("Announcing %s to master", config["agent_hostname"])

        def callback(response):
            if response.code == OK:
                self.reannounce_client_request = None
                config.master_contacted(announcement=True)
                svclog.info("Announced self to the master server.")

            elif response.code >= INTERNAL_SERVER_ERROR:
                if not self.shutting_down:
                    delay = http_retry_delay()
                    svclog.warning(
                        "Could not announce self to the master server, "
                        "internal server error: %s.  Retrying in %s seconds.",
                        response.data(), delay)
                    reactor.callLater(delay, response.request.retry)
                else:
                    svclog.warning("Could not announce to master. Not retrying "
                                   "because of pending shutdown")

            # If this is a client problem retrying the request
            # is unlikely to fix the issue so we stop here
            elif response.code >= BAD_REQUEST:
                self.reannounce_client_request = None
                svclog.error(
                    "Failed to announce self to the master, bad "
                    "request: %s.  This request will not be retried.",
                    response.data())

            else:
                self.reannounce_client_request = None
                svclog.error(
                    "Unhandled error when posting self to the "
                    "master: %s (code: %s).  This request will not be "
                    "retried.", response.data(), response.code)

        # In the event of a hard failure, do not retry because we'll
        # be rerunning this code soon anyway.
        def errback(failure):
            self.reannounce_client_request = None
            svclog.error(
                "Failed to announce self to the master: %s.  This "
                "request will not be retried.", failure)

        self.reannounce_client_request = post(
            self.agent_api(),
            callback=callback, errback=errback,
            data={
                "state": config["state"],
                "current_assignments": config.get(
                    "current_assignments", {}),  # may not be set yet
                "free_ram": memory.free_ram()})

    def system_data(self, requery_timeoffset=False):
        """
        Returns a dictionary of data containing information about the
        agent.  This is the information that is also passed along to
        the master.
        """
        # query the time offset and then cache it since
        # this is typically a blocking operation
        if config["agent_time_offset"] == "auto":
            config["agent_time_offset"] = None

        if requery_timeoffset or config["agent_time_offset"] is None:
            ntplog.info(
                "Querying ntp server %r for current time",
                config["agent_ntp_server"])

            ntp_client = NTPClient()
            try:
                pool_time = ntp_client.request(
                    config["agent_ntp_server"],
                    version=config["agent_ntp_server_version"])

            except Exception as e:
                ntplog.warning("Failed to determine network time: %s", e)

            else:
                config["agent_time_offset"] = \
                    int(pool_time.tx_time - time.time())

                # format the offset for logging purposes
                utcoffset = datetime.utcfromtimestamp(pool_time.tx_time)
                iso_timestamp = utcoffset.isoformat()
                ntplog.debug(
                    "network time: %s (local offset: %r)",
                    iso_timestamp, config["agent_time_offset"])

                if config["agent_time_offset"] != 0:
                    ntplog.warning(
                        "Agent is %r second(s) off from ntp server at %r",
                        config["agent_time_offset"],
                        config["agent_ntp_server"])

        data = {
            "systemid": config["agent_systemid"],
            "hostname": config["agent_hostname"],
            "version": config.version,
            "ram": int(config["agent_ram"]),
            "cpus": config["agent_cpus"],
            "port": config["agent_api_port"],
            "free_ram": memory.free_ram(),
            "time_offset": config["agent_time_offset"] or 0,
            "state": config["state"],
            "current_assignments": config.get(
                "current_assignments", {})}  # may not be set yet

        if "remote_ip" in config:
            data.update(remote_ip=config["remote_ip"])

        if "projects" in config:
           data.update(projects=config["projects"])

        return data

    def build_http_resource(self):
        svclog.debug("Building HTTP Service")
        root = Resource()

        # static endpoints to redirect resources
        # to the right objects
        root.putChild(
            "favicon.ico",
            StaticPath(join(config["agent_static_root"], "favicon.ico"),
                       defaultType="image/x-icon"))
        root.putChild(
            "static",
            StaticPath(config["agent_static_root"]))

        # external endpoints
        root.putChild("", Index())
        root.putChild("configuration", Configuration())

        # api endpoints
        api = root.putChild("api", APIRoot())
        api.putChild("versions", Versions())
        v1 = api.putChild("v1", APIRoot())

        # Top level api endpoints
        v1.putChild("assign", Assign(self))
        v1.putChild("tasks", Tasks())
        v1.putChild("config", Config())
        v1.putChild("logging", LogQuery())
        v1.putChild("task_logs", TaskLogs())

        # Endpoints which are generally used for status
        # and operations.
        v1.putChild("status", Status())
        v1.putChild("stop", Stop())
        v1.putChild("update", Update())

        return root

    def _start_manhole(self, port, username, password):
        """
        Starts the manhole server so we can connect to the agent
        over telnet
        """
        if "manhole" in self.services:
            svclog.warning(
                "Telnet manhole service is already running on port %s",
                self.services["manhole"].port)
            return

        svclog.info("Starting telnet manhole on port %s", port)

        # Since we don't always need this module we import
        # it here to save on memory and other resources
        from pyfarm.agent.manhole import manhole_factory

        # Contains the things which will be in the top level
        # namespace of the Python interpreter.
        namespace = {
            "config": config, "agent": self,
            "jobtypes": config["jobtypes"],
            "current_assignments": config["current_assignments"]}

        factory = manhole_factory(namespace, username, password)
        self.services["manhole"] = reactor.listenTCP(port, factory)

    def _start_http_api(self, port):
        """
        Starts the HTTP api so the master can communicate
        with the agent.
        """
        if "api" in self.services:
            svclog.warning(
                "HTTP API service already running on port %s",
                self.services["api"].port)
            return

        http_resource = self.build_http_resource()
        self.services["api"] = reactor.listenTCP(port, Site(http_resource))

    def start(self, shutdown_events=True, http_server=True):
        """
        Internal code which starts the agent, registers it with the master,
        and performs the other steps necessary to get things running.

        :param bool shutdown_events:
            If True register all shutdown events so certain actions, such as
            information the master we're going offline, can take place.

        :param bool http_server:
            If True then construct and serve the externally facing http
            server.
        """
        if config["agent_manhole"]:
            self._start_manhole(config["agent_manhole_port"],
                                config["agent_manhole_username"],
                                config["agent_manhole_password"])

        # setup the internal http server so external entities can
        # interact with the service.
        if http_server:
            self._start_http_api(config["agent_api_port"])

        # Update the configuration with this pid (which may be different
        # than the original pid).
        config["pids"].update(child=os.getpid())

        # get ready to 'publish' the agent
        config.register_callback(
            "agent-id",
            partial(
                self.callback_agent_id_set, shutdown_events=shutdown_events))
        config.register_callback("free_ram", self.callback_free_ram_changed)
        config.register_callback("cpus", self.callback_cpu_count_changed)
        return self.post_agent_to_master()

    def stop(self):
        """
        Internal code which stops the agent.  This will terminate any running
        processes, inform the master of the terminated tasks, update the
        state of the agent on the master.
        """
        svclog.info("Stopping the agent")

        # If this is the first time we're calling stop() then
        # setup a function call to remove the pidfile when the
        # process exits.
        if not self.shutting_down:
            self.shutting_down = True
            self.shutdown_timeout = (
                datetime.utcnow() + timedelta(
                    seconds=config["agent_shutdown_timeout"]))

            def remove_pidfile():
                if not isfile(config["agent_lock_file"]):
                    svclog.warning(
                        "%s does not exist", config["agent_lock_file"])
                    return

                try:
                    os.remove(config["agent_lock_file"])
                    svclog.debug(
                        "Removed pidfile %r", config["agent_lock_file"])
                except (OSError, IOError) as e:
                    svclog.error(
                        "Failed to remove lock file %r: %s",
                        config["agent_lock_file"], e)

            atexit.register(remove_pidfile)

            self.scheduled_tasks.stop()

            svclog.debug("Stopping execution of jobtypes")
            for uuid, jobtype in config["jobtypes"].items():
                jobtype.stop()

            def wait_on_stopped():
                svclog.info("Waiting on %s job types to terminate",
                            len(config["jobtypes"]))

                for jobtype_id, jobtype in config["jobtypes"].copy().items():
                    if not jobtype._has_running_processes():
                        svclog.error(
                            "%r has not removed itself, forcing removal",
                            jobtype)
                        config["jobtypes"].pop(jobtype_id)

                if config["jobtypes"]:
                    reactor.callLater(1, wait_on_stopped)
                elif self.agent_api() is not None:
                    self.post_shutdown_to_master()
                else:
                    reactor.stop()

            wait_on_stopped()

    def sigint_handler(self, *_):
        self.stop()

    def post_shutdown_to_master(self, stop_reactor=True):
        """
        This method is called before the reactor shuts down and lets the
        master know that the agent's state is now ``offline``
        """
        svclog.info("Informing master of shutdown")

        # This deferred is fired when we've either been successful
        # or failed to letting the master know we're shutting
        # down.  It is the last object to fire so if you have to do
        # other things, like stop the process, that should happen before
        # the callback for this one is run.
        finished = Deferred()

        def post_update(run=True):
            # NOTE: current_assignments *should* be empty now because the tasks
            # should shutdown first.  If not then we'll let the master know.
            def perform():
                return post(
                    self.agent_api(),
                    data={
                        "state": AgentState.OFFLINE,
                        "free_ram": memory.free_ram(),
                        "current_assignments": config["current_assignments"]},
                    callback=results_from_post,
                    errback=error_while_posting)

            return perform() if run else perform

        def results_from_post(response):
            if response.code == NOT_FOUND:
                svclog.warning(
                    "Agent %r no longer exists, cannot update state",
                    config["agent-id"])
                finished.callback(NOT_FOUND)

            elif response.code == OK:
                svclog.info(
                    "Agent %r has shutdown successfully", config["agent-id"])
                finished.callback(OK)

            elif response.code >= INTERNAL_SERVER_ERROR:
                if self.shutdown_timeout > datetime.utcnow():
                    delay = http_retry_delay()
                    svclog.warning(
                        "State update failed due to server error: %s.  "
                        "Retrying in %s seconds.",
                        response.data(), delay)
                    reactor.callLater(delay, response.request.retry)
                else:
                    svclog.warning(
                        "State update failed due to server error: %s.  "
                        "Shutdown timeout reached, not retrying.",
                        response.data())
                    finished.errback(INTERNAL_SERVER_ERROR)

            else:
                if self.shutdown_timeout > datetime.utcnow():
                    delay = http_retry_delay()
                    svclog.warning(
                        "State update failed due to unhandled error: %s.  "
                        "Retrying in %s seconds.",
                        response.data(), delay)
                    reactor.callLater(delay, response.request.retry)
                else:
                    svclog.warning(
                        "State update failed due to unhandled error: %s.  "
                        "Shutdown timeout reached, not retrying.",
                        response.data())
                    finished.errback(response.code)

        def error_while_posting(failure):
            if self.shutdown_timeout > datetime.utcnow():
                delay = http_retry_delay()
                svclog.warning(
                    "State update failed due to unhandled error: %s.  "
                    "Retrying in %s seconds",
                    failure, delay)
                reactor.callLater(delay, post_update(run=False))
            else:
                svclog.warning(
                    "State update failed due to unhandled error: %s.  "
                    "Shutdown timeout reached, not retrying.",
                    failure)
                finished.errback(failure)

        # Post our current state to the master.  We're only posting free_ram
        # and state here because all other fields would be updated the next
        # time the agent starts up.  free_ram would be too but having it
        # here is beneficial in cases where the agent terminated abnormally.
        post_update()

        if stop_reactor:
            finished.addBoth(lambda *_: reactor.stop())

        return finished

    def errback_post_agent_to_master(self, failure):
        """
        Called when there's a failure trying to post the agent to the
        master.  This is often because of some lower level issue but it
        may be recoverable to we retry the request.
        """
        delay = http_retry_delay()

        if failure.type is ConnectionRefusedError and not self.shutting_down:
            svclog.error(
                "Failed to POST agent to master, the connection was refused. "
                "Retrying in %s seconds", delay)
        elif not self.shutting_down:
            svclog.error(
                "Unhandled error when trying to POST the agent to the master. "
                "The error was %s.  Retrying in %s seconds.", failure, delay)
        else:
            svclog.error(
                "Failed to POST agent to master. Not retrying, because the "
                "agent is shutting down.")

        if not self.shutting_down:
            return reactor.callLater(
                http_retry_delay(), self.post_agent_to_master)

    def callback_post_agent_to_master(self, response):
        """
        Called when we get a response after POSTing the agent to the
        master.
        """
        # Master might be down or some other internal problem
        # that might eventually be fixed.  Retry the request.
        if response.code >= 500:
            if not self.shutting_down:
                delay = http_retry_delay()
                svclog.warning(
                    "Failed to post to master due to a server side error "
                    "error %s, retrying in %s seconds", response.code, delay)
                reactor.callLater(delay, self.post_agent_to_master)
            else:
                svclog.warning(
                    "Failed to post to master due to a server side error "
                    "error %s. Not retrying, because the agent is shutting down",
                    response.code)

        # Master is up but is rejecting our request because there's something
        # wrong with it.  Do not retry the request.
        elif response.code >= 400:
            svclog.error(
                "%s accepted our POST request but responded with code %s "
                "which is a client side error.  The message the server "
                "responded with was %r.  Sorry, but we cannot retry this "
                "request as it's an issue with the agent's request.",
                self.agents_endpoint(), response.code, response.data())

        else:
            data = response.json()
            config["agent-id"] = data["id"]
            config.master_contacted()

            if response.code == OK:
                svclog.info(
                    "POST to %s was successful. Agent %s was updated.",
                    self.agents_endpoint(), config["agent-id"])

            elif response.code == CREATED:
                svclog.info(
                    "POST to %s was successful.  A new agent "
                    "with an id of %s was created.",
                    self.agents_endpoint(), config["agent-id"])

    def post_agent_to_master(self):
        """
        Runs the POST request to contact the master.  Running this method
        multiple times should be considered safe but is generally something
        that should be avoided.
        """
        return post(
            self.agents_endpoint(),
            callback=self.callback_post_agent_to_master,
            errback=self.errback_post_agent_to_master,
            data=self.system_data())

    def callback_post_free_ram(self, response):
        """
        Called when we get a response back from the master
        after POSTing a change for ``free_ram``
        """
        self.last_free_ram_post = time.time()
        if response.code == OK:
            svclog.info(
                "POST %s {'free_ram': %d}` succeeded",
                self.agent_api(), response.json()["free_ram"])

        # Because this happens with a fairly high frequency we don't
        # retry failed requests because we'll be running `post_free_ram`
        # soon again anyway
        else:
            svclog.warning(
                "Failed to post `free_ram` to %s: %s %s - %s",
                self.agent_api(), response.code, responses[response.code],
                response.json())

    def errback_post_free_ram(self, failure):
        """
        Error handler which is called if we fail to post a ram
        update to the master for some reason
        """
        svclog.error(
            "Failed to post update to `free_ram` to the master: %s", failure)

    def post_free_ram(self):
        """
        Posts the current nu
        """
        since_last_update = time.time() - self.last_free_ram_post
        left_till_update = \
            config["agent_ram_max_report_frequency"] - since_last_update

        if left_till_update > 0:
            svclog.debug(
                "Skipping POST for `free_ram` change, %.2f seconds left "
                "in current interval.", left_till_update)
            deferred = Deferred()
            deferred.callback("delay")
            return deferred
        else:
            return post(self.agent_api(),
                data={"free_ram": config["free_ram"]},
                callback=self.callback_post_free_ram,
                errback=self.errback_post_free_ram)

    def callback_free_ram_changed(self, change_type, key, new_value, old_value):
        """
        Callback used to decide and act on changes to the
        ``config['ram']`` value.
        """
        if change_type == config.MODIFIED:
            if abs(new_value - old_value) < config["agent_ram_report_delta"]:
                svclog.debug("Not enough change in free_ram to report")
            else:
                self.post_free_ram()

    def errback_post_cpu_count_change(self, failure):
        """
        Error handler which is called if we fail to post a cpu count update
        to an existing agent for some reason.
        """
        delay = http_retry_delay()
        svclog.warning(
            "There was error updating an existing agent: %s.  Retrying "
            "in %r seconds", failure, delay)
        reactor.callLater(delay, self.post_cpu_count(run=False))

    def callback_post_cpu_count_change(self, response):
        """
        Called when we received a response from the master after
        """
        if response.code == OK:
            svclog.info("CPU count change POSTed to %s", self.agent_api())
        else:
            delay = http_retry_delay()
            svclog.warning(
                "We expected to receive an OK response code but instead"
                "we got %s.  Retrying in %s.", responses[response.code], delay)
            reactor.callLater(delay, self.post_cpu_count(run=False))

    def post_cpu_count(self, run=True):
        """POSTs CPU count changes to the master"""
        def run_post():
            return post(self.agent_api(),
                data={"cpus": config["agent_cpus"]},
                callback=self.callback_post_cpu_count_change,
                errback=self.errback_post_cpu_count_change)
        return run_post() if run else run_post

    def callback_cpu_count_changed(
            self, change_type, key, new_value, old_value):
        """
        Callback used to decide and act on changes to the
        ``config['cpus']`` value.
        """
        if change_type == config.MODIFIED and new_value != old_value:
            svclog.debug(
                "CPU count has been changed from %s to %s",
                old_value, new_value)
            self.post_cpu_count()

    def callback_agent_id_set(
            self, change_type, key, new_value, old_value, shutdown_events=True):
        """
        When `agent-id` is created we need to:

            * Register a shutdown event so that when the agent is told to
              shutdown it will notify the master of a state change.
            * Star the scheduled task manager
        """
        if key == "agent-id" and change_type == config.CREATED \
                and not self.register_shutdown_events:
            if shutdown_events:
                self.register_shutdown_events = True

            # set the initial free_ram
            config["free_ram"] = memory.free_ram()

            config.master_contacted()
            svclog.debug(
                "`%s` was %s, adding system event trigger for shutdown",
                key, change_type)

            self.scheduled_tasks.start()



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from datetime import timedelta, datetime

try:
    from httplib import ACCEPTED, OK, BAD_REQUEST
except ImportError:  # pragma: no cover
    from http.client import ACCEPTED, OK, BAD_REQUEST

import psutil
from twisted.internet.defer import Deferred
from twisted.web.server import NOT_DONE_YET
from voluptuous import Schema, Optional

from pyfarm.agent.config import config
from pyfarm.agent.http.api.base import APIResource
from pyfarm.agent.logger import getLogger
from pyfarm.agent.sysinfo import memory
from pyfarm.agent.utility import dumps, total_seconds

logger = getLogger("agent.http.state")


class Stop(APIResource):
    isLeaf = False  # this is not really a collection of things
    SCHEMAS = {
        "POST": Schema({
            Optional("wait"): bool})}

    def post(self, **kwargs):
        request = kwargs["request"]
        data = kwargs["data"]
        agent = config["agent"]
        stopping = agent.stop()

        # TODO: need to wire this up to the real deferred object in stop()
        if data.get("wait"):
            def finished(_, finish_request):
                finish_request.setResponseCode(OK)
                finish_request.finish()

            if isinstance(stopping, Deferred):
                stopping.addCallback(finished, request)
            else:  # pragma: no cover
                request.setResponseCode(OK)
                request.finish()

        else:
            request.setResponseCode(ACCEPTED)
            request.finish()

        return NOT_DONE_YET


class Status(APIResource):
    isLeaf = False  # this is not really a collection of things

    def get(self, **_):
        # Get counts for child processes and grandchild processes
        process = psutil.Process()
        direct_child_processes = len(process.children(recursive=False))
        all_child_processes = len(process.children(recursive=True))
        grandchild_processes = all_child_processes - direct_child_processes

        # Determine the last time we talked to the master (if ever)
        contacted = config.master_contacted(update=False)
        if isinstance(contacted, datetime):  # pragma: no cover
            contacted = datetime.utcnow() - contacted

        # Determine the last time we announced ourselves to the
        # master (if ever)
        last_announce = config.get("last_announce", None)
        if isinstance(last_announce, datetime):  # pragma: no cover
            last_announce = datetime.utcnow() - last_announce

        return dumps(
            {"state": config["state"],
             "agent_hostname": config["agent_hostname"],
             "free_ram": memory.free_ram(),
             "agent_process_ram": memory.process_memory(),
             "consumed_ram": memory.total_consumption(),
             "child_processes": direct_child_processes,
             "grandchild_processes": grandchild_processes,
             "pids": config["pids"],
             "id": config.get("agent-id", None),
             "agent_systemid": config["agent_systemid"],
             "last_master_contact": contacted,
             "last_announce": last_announce,
             "agent_lock_file": config["agent_lock_file"],
             "uptime": total_seconds(
                 timedelta(seconds=time.time() - config["start"])),
             "jobs": list(config["jobtypes"].keys())})



# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. |ProcessProtocol| replace:: pyfarm.jobtypes.core.process.ProcessProtocol

Job Type Core
=============

This module contains the core job type from which all
other job types are built.  All other job types must
inherit from the :class:`JobType` class in this modle.
"""

import os
import tempfile
from errno import EEXIST
from datetime import datetime, timedelta
from functools import partial
from string import Template
from os.path import expanduser, abspath, isdir, join
from pprint import pformat

try:
    from httplib import OK, INTERNAL_SERVER_ERROR, CONFLICT, CREATED
except ImportError:  # pragma: no cover
    from http.client import OK, INTERNAL_SERVER_ERROR, CONFLICT, CREATED

from twisted.internet import reactor
from twisted.internet.error import ProcessDone, ProcessTerminated
from twisted.python.failure import Failure
from voluptuous import Schema, Required, Optional

from pyfarm.core.enums import INTEGER_TYPES, STRING_TYPES, WorkState, WINDOWS
from pyfarm.core.utility import ImmutableDict
from pyfarm.agent.config import config
from pyfarm.agent.http.core.client import post, http_retry_delay
from pyfarm.agent.logger import getLogger
from pyfarm.agent.sysinfo import memory, system
from pyfarm.agent.sysinfo.user import is_administrator, username
from pyfarm.agent.utility import (
    TASKS_SCHEMA, JOBTYPE_SCHEMA, JOB_SCHEMA, uuid, validate_uuid)
from pyfarm.jobtypes.core.internals import (
    USER_GROUP_TYPES, Cache, Process, TypeChecks, System, pwd, grp)
from pyfarm.jobtypes.core.log import STDOUT, STDERR, logpool
from pyfarm.jobtypes.core.process import ProcessProtocol

logcache = getLogger("jobtypes.cache")
logger = getLogger("jobtypes.core")
process_stdout = getLogger("jobtypes.process.stdout")
process_stderr = getLogger("jobtypes.process.stderr")


FROZEN_ENVIRONMENT = ImmutableDict(os.environ.copy())


class CommandData(object):
    """
    Stores data to be returned by :meth:`JobType.get_command_data`.  Instances
    of this class are alosed used by :meth:`JobType.spawn_process_inputs` at
    execution time.

    .. note::

        This class does not perform any key of path resolution by default.  It
        is assumed this has already been done using something like
        :meth:`JobType.map_path`

    :arg string command:
        The command that will be executed when the process runs.

    :arg arguments:
        Any additional arguments to be passed along to the command being
        launched.

    :keyword dict env:
        If provided, this will be the environment to launch the command
        with.  If this value is not provided then a default environment
        will be setup using :meth:`set_default_environment` when
        :meth:`JobType.start` is called.  :meth:`JobType.start` itself will
        use :meth:`JobType.default_environment` to generate the default
        environment.

    :keyword string cwd:
        The working directory the process should execute in.  If not provided
        the process will execute in whatever the directory the agent is
        running inside of.

    :type user: string or integer
    :keyword user:
        The username or user id that the process should run as.  On Windows
        this keyword is ignored and on Linux this requires the agent to be
        executing as root.  The value provided here will be run through
        :meth:`JobType.get_uid_gid` to map the incoming value to an integer.

    :type group: string or integer
    :keyword group:
        Same as ``user`` above except this sets the group the process will
        execute.
    """
    def __init__(self, command, *arguments, **kwargs):
        self.command = command
        self.arguments = tuple(map(str, arguments))
        self.env = kwargs.pop("env", {})
        self.cwd = kwargs.pop("cwd", None)
        self.user = kwargs.pop("user", None)
        self.group = kwargs.pop("group", None)

        if kwargs:
            raise ValueError(
                "Unexpected keywords present in kwargs: %s" % kwargs.keys())

    def __repr__(self):
        return "Command(command=%r, arguments=%r, cwd=%r, user=%r, group=%r, " \
               "env=%r)" % (self.command, self.arguments, self.cwd,
                            self.user, self.group, self.env)

    def validate(self):
        """
        Validates that the attributes on an instance of this class contain
        values we expect.  This method is called externally by the job type in
        :meth:`JobType.start` and may correct some instance attributes.
        """
        if not isinstance(self.command, STRING_TYPES):
            raise TypeError("Expected a string for `command`")

        if not isinstance(self.env, dict):
            raise TypeError("Expected a dictionary for `env`")

        if not isinstance(self.user, USER_GROUP_TYPES):
            raise TypeError("Expected string, integer or None for `user`")

        if not isinstance(self.group, USER_GROUP_TYPES):
            raise TypeError("Expected string, integer or None for `group`")

        if WINDOWS:  # pragma: no cover
            if self.user is not None:
                logger.warning("`user` is ignored on Windows")
                self.user = None

            if self.group is not None:
                logger.warning("`group` is ignored on Windows")
                self.group = None

        elif self.user is not None or self.group is not None \
                and not is_administrator():
            raise EnvironmentError(
                "Cannot change user or group without being admin.")

        if self.cwd is None:
            if not config["agent_chdir"]:
                self.cwd = os.getcwd()
            else:
                self.cwd = config["agent_chdir"]

        if isinstance(self.cwd, STRING_TYPES):
            if not isdir(self.cwd):
                raise OSError(
                    "`cwd` %s does not exist" % self.cwd)

        elif self.cwd is not None:
            raise TypeError("Expected a string for `cwd`")

    def set_default_environment(self, value):
        """
        Sets the environment to ``value`` if the internal ``env`` attribute
        is None.  By default this method is called by the job type and passed
        in the results from :meth:`pyfarm.jobtype.core.JobType.get_environment`
        """
        if self.env is None:
            assert isinstance(value, dict)
            self.env = value


class JobType(Cache, System, Process, TypeChecks):
    """
    Base class for all other job types.  This class is intended
    to abstract away many of the asynchronous necessary to run
    a job type on an agent.

    :cvar set PERSISTENT_JOB_DATA:
        A dictionary of job ids and data that :meth:`prepare_for_job` has
        produced.  This is used during :meth:`__init__` to set
        ``persistent_job_data``.

    :cvar CommandData COMMAND_DATA_CLASS:
        If you need to provide your own class to represent command data you
        should override this attribute.  This attribute is used by by methods
        within this class to do type checking.

    :cvar ProcessProtocol PROCESS_PROTOCOL:
        The protocol object used to communicate with each process
        spawned

    :cvar voluptuous.Schema ASSIGNMENT_SCHEMA:
        The schema of an assignment.  This object helps to validate the
        incoming assignment to ensure it's not missing any data.

    :arg dict assignment:
        This attribute is a dictionary the keys "job", "jobtype" and "tasks".
        self.assignment["job"] is itself a dict with keys "id", "title",
        "data", "environ" and "by".  The most important of those is usually
        "data", which is the dict specified when submitting the job and
        contains jobtype specific data. self.assignment["tasks"] is a list of
        dicts representing the tasks in the current assignment.  Each of
        these dicts has the keys "id" and "frame".  The
        list is ordered by frame number.

    :ivar UUID uuid:
        This is the unique identifier for the job type instance and is
        automatically set when the class is instanced.  This is used by the
        agent to track assignments and job type instances.

    :ivar set finished_tasks:
        A set of tasks that have had their state changed to finished through
        :meth:`set_task_state`.  At the start of the assignment, this list is
        empty.

    :ivar set failed_tasks:
        This is analogous to ``finished_tasks`` except it contains failed
        tasks only.
    """
    PERSISTENT_JOB_DATA = {}
    COMMAND_DATA = CommandData
    PROCESS_PROTOCOL = ProcessProtocol
    ASSIGNMENT_SCHEMA = Schema({
        Required("id"): validate_uuid,
        Required("job"): JOB_SCHEMA,
        Required("jobtype"): JOBTYPE_SCHEMA,
        Optional("tasks"): TASKS_SCHEMA})

    def __init__(self, assignment):
        super(JobType, self).__init__()

        # Private attributes which persist with the instance.  These
        # generally should not be modified directly.
        self._tempdir = None  # the defacto tempdir for this instance
        self._tempdirs = set()  # list of any temp directories created
        self._stdout_line_fragments = {}
        self._stderr_line_fragments = {}

        # JobType objects in the future may or may not have explicit tasks
        # associated with when them.  The format of tasks could also change
        # since it's an internal representation so to guard against these
        # changes we just use a simple uuid to represent ourselves.
        self.uuid = uuid()
        self.processes = {}
        self.failed_processes = set()
        self.failed_tasks = set()
        self.finished_tasks = set()
        self.assignment = ImmutableDict(self.ASSIGNMENT_SCHEMA(assignment))
        self.persistent_job_data = None

        # Add our instance to the job type instance tracker dictionary
        # as well as the dictionary containing the current assignment.
        config["jobtypes"][self.uuid] = self
        config["current_assignments"][assignment["id"]]["jobtype"].update(
            id=self.uuid)

        # NOTE: Don't call this logging statement before the above, we need
        # self.assignment
        logger.debug("Instanced %r", self)

    def __repr__(self):
        formatting = "%s(job=%r, tasks=%r, jobtype=%r, version=%r, title=%r)"
        return formatting % (
            self.__class__.__name__,
            self.assignment["job"]["id"],
            tuple(task["id"] for task in self.assignment["tasks"]),
            str(self.assignment["jobtype"]["name"]),
            self.assignment["jobtype"]["version"],
            str(self.assignment["job"]["title"]))

    @classmethod
    def load(cls, assignment):
        """
        Given an assignment this class method will load the job type either
        from cache or from the master.

        :param dict assignment:
            The dictionary containing the assignment.  This will be
            passed into an instance of ``ASSIGNMENT_SCHEMA`` to validate
            that the internal data is correct.
        """
        cls.ASSIGNMENT_SCHEMA(assignment)

        cache_key = cls._cache_key(assignment)
        logger.debug("Cache key for assignment is %s", cache_key)

        if config["jobtype_enable_cache"] or cache_key not in cls.cache:
            logger.debug("Jobtype not in cache or cache disabled")
            download = cls._download_jobtype(
                assignment["jobtype"]["name"],
                assignment["jobtype"]["version"])
            download.addCallback(cls._jobtype_download_complete, cache_key)
            return download
        else:
            logger.debug("Caching jobtype")
            return cls._load_jobtype(cls.cache[cache_key], None)

    @classmethod
    def prepare_for_job(cls, job):
        """
        .. note::
            This method is not yet implemented

        Called before a job executes on the agent first the first time.
        Whatever this classmethod returns will be available as
        ``persistent_job_data`` on the job type instance.

        :param int job:
            The job id which prepare_for_job is being run for

        By default this method does nothing.
        """
        pass

    @classmethod
    def cleanup_after_job(cls, persistent_data):
        """
        .. note::
            This method is not yet implemented

        This classmethod will be called after the last assignment
        from a given job has finished on this node.

        :param persistent_data:
            The persistent data that :meth:`prepare_for_job` produced.  The
            value for this data may be ``None`` if :meth:`prepare_for_job`
            returned None or was not implemented.
        """
        pass

    @classmethod
    def spawn_persistent_process(cls, job, command_data):
        """
        .. note::
            This method is not yet implemented

        Starts one child process using an instance of :class:`CommandData` or
        similiar input.  This process is intended to keep running until the
        last task from this job has been processed, potentially spanning more
        than one assignment.  If the spawned process is still running then
        we'll cleanup the process after :meth:`cleanup_after_job`
        """
        pass

    def node(self):
        """
        Returns live information about this host, the operating system,
        hardware, and several other pieces of global data which is useful
        inside of the job type.  Currently data from this method includes:

            * **master_api** - The base url the agent is using to
              communicate with the master.
            * **hostname** - The hostname as reported to the master.
            * **systemid** - The unique identifier used to identify.
              this agent to the master.
            * **id** - The database id of the agent as given to us by
              the master on startup of the agent.
            * **cpus** - The number of CPUs reported to the master
            * **ram** - The amount of ram reported to the master.
            * **total_ram** - The amount of ram, in megabytes,
              that's installed on the system regardless of what
              was reported to the master.
            * **free_ram** - How much ram, in megabytes, is free
              for the entire system.
            * **consumed_ram** - How much ram, in megabytes, is
              being consumed by the agent and any processes it has
              launched.
            * **admin** - Set to True if the current user is an
              administrator or 'root'.
            * **user** - The username of the current user.
            * **case_sensitive_files** - True if the file system is
              case sensitive.
            * **case_sensitive_env** - True if environment variables
              are case sensitive.
            * **machine_architecture** - The architecture of the machine
              the agent is running on.  This will return 32 or 64.
            * **operating_system** - The operating system the agent
              is executing on.  This value will be 'linux', 'mac' or
              'windows'.  In rare circumstances this could also
              be 'other'.

        :raises KeyError:
            Raised if one or more keys are not present in
            the global configuration object.

            This should rarely if ever be a problem under normal
            circumstances.  The exception to this rule is in
            unittests or standalone libraries with the global
            config object may not be populated.
        """
        try:
            machine_architecture = system.machine_architecture()
        except NotImplementedError:
            logger.warning(
                "Failed to determine machine architecture.  This is a "
                "bug, please report it.")
            raise

        return {
            "master_api": config.get("master-api"),
            "hostname": config["agent_hostname"],
            "systemid": config["agent_systemid"],
            "id": int(config["agent-id"]),
            "cpus": int(config["agent_cpus"]),
            "ram": int(config["agent_ram"]),
            "total_ram": int(memory.total_ram()),
            "free_ram": int(memory.free_ram()),
            "consumed_ram": int(memory.total_consumption()),
            "admin": is_administrator(),
            "user": username(),
            "case_sensitive_files": system.filesystem_is_case_sensitive(),
            "case_sensitive_env": system.environment_is_case_sensitive(),
            "machine_architecture": machine_architecture,
            "operating_system": system.operating_system()}

    def assignments(self):
        """Short cut method to access tasks"""
        return self.assignment["tasks"]

    def tempdir(self, new=False, remove_on_finish=True):
        """
        Returns a temporary directory to be used within a job type.
        By default once called the directory will be created on disk
        and returned from this method.

        Calling this method multiple times will return the same directory
        instead of creating a new directory unless ``new`` is set to True.

        :param bool new:
            If set to ``True`` then return a new directory when called.  This
            however will not replace the 'default' temp directory.

        :param bool remove_on_finish:
            If ``True`` then keep track of the directory returned so it
            can be removed when the job type finishes.
        """
        if not new and self._tempdir is not None:
            return self._tempdir

        parent_dir = config["jobtype_tempdir_root"].replace(
            "$JOBTYPE_UUID", str(self.uuid))

        try:
            os.makedirs(parent_dir)
        except OSError as e:  # pragma: no cover
            if e.errno != EEXIST:
                logger.error("Failed to create %s: %s", parent_dir, e)
                raise

        self._tempdirs.add(parent_dir)
        tempdir = tempfile.mkdtemp(dir=parent_dir)
        logger.debug(
            "%s.tempdir() created %s", self.__class__.__name__, tempdir)

        # Keep track of the directory so we can cleanup all of them
        # when the job type finishes.
        if remove_on_finish:
            self._tempdirs.add(tempdir)

        if not new and self._tempdir is None:
            self._tempdir = tempdir

        return tempdir

    def get_uid_gid(self, user, group):
        """
        **Overridable**.  This method to convert a named user and group
        into their respective user and group ids.
        """
        uid = None
        gid = None

        # Convert user to uid
        if all([user is not None, pwd is not NotImplemented]):
            uid = self._get_uid_gid_value(
                user, "username", "get_uid", pwd, "pwd")

        # Convert group to gid
        if all([group is not None, grp is not NotImplemented]):
            gid = self._get_uid_gid_value(
                group, "group", "get_gid", grp, "grp")

        return uid, gid

    def get_environment(self):
        """
        Constructs an environment dictionary that can be used
        when a process is spawned by a job type.
        """
        environment = {}
        config_environment = config.get("jobtype_default_environment")

        if config.get("jobtype_include_os_environ"):
            environment.update(FROZEN_ENVIRONMENT)

        if isinstance(config_environment, dict):
            environment.update(config_environment)

        elif config_environment is not None:
            logger.warning(
                "Expected a dictionary for `jobtype_default_environment`, "
                "ignoring the configuration value.")

        # All keys and values must be strings.  Normally this is not an issue
        # but it's possible to create an environment using the config which
        # contains non-strings.
        for key in environment:
            if not isinstance(key, STRING_TYPES):
                logger.warning(
                    "Environment key %r is not a string.  It will be converted "
                    "to a string.", key)

                value = environment.pop(key)
                key = str(key)
                environment[key] = value

            if not isinstance(environment[key], STRING_TYPES):
                logger.warning(
                    "Environment value for %r is not a string.  It will be "
                    "converted to a string.", key)
                environment[key] = str(environment[key])

        return environment

    def get_command_list(self, cmdlist):
        """
        Return a list of command to be used when running the process
        as a read-only tuple.
        """
        self._check_command_list_inputs(cmdlist)
        return tuple(map(self.expandvars, cmdlist))

    def get_csvlog_path(self, protocol_uuid, create_time=None):
        """
        Returns the path to the comma separated value (csv) log file.
        The agent stores logs from processes in a csv format so we can store
        additional information such as a timestamp, line number, stdout/stderr
        identification and the the log message itself.

        .. note::

            This method should not attempt to create the parent directories
            of the resulting path.  This is already handled by the logger
            pool in a non-blocking fashion.
        """
        self._check_csvlog_path_inputs(protocol_uuid, create_time)

        if create_time is None:
            create_time = datetime.utcnow()
        assert isinstance(create_time, datetime)

        # Include the agent's time offset in create_time for accuracy.
        if config["agent_time_offset"]:
            create_time += timedelta(seconds=config["agent_time_offset"])

        # The default string template implementation cannot
        # handle cases where you have $VARS$LIKE_$THIS.  So we
        # instead iterate over a dictionary and use .replace()
        template_data = {
            "YEAR": str(create_time.year),
            "MONTH": "%02d" % create_time.month,
            "DAY": "%02d" % create_time.day,
            "HOUR": "%02d" % create_time.hour,
            "MINUTE": "%02d" % create_time.minute,
            "SECOND": "%02d" % create_time.second,
            "JOB": str(self.assignment["job"]["id"]),
            "PROCESS": protocol_uuid.hex}

        filename = config["jobtype_task_log_filename"]
        for key, value in template_data.items():
            filename = filename.replace("$" + key, value)
        filepath = join(config["jobtype_task_logs"], filename)

        return abspath(filepath)

    def get_command_data(self):
        """
        **Overridable**.  This method returns the arguments necessary for
        executing a command.  For job types which execute a single process
        per assignment, this is the most important method to implement.

        .. warning::

            This method should not be used when this jobtype requires more
            than one process for one assignment and may not get called at all
            if :meth:`start` was overridden.

        The default implementation does nothing.  When overriding this method
        you should return an instance of ``COMMAND_DATA_CLASS``:

        .. code-block:: python

            return self.COMMAND_DATA(
                "/usr/bin/python", "-c", "print 'hello world'",
                env={"FOO": "bar"}, user="bob")

        See :class:`CommandData`'s class documentation for a full description
        of possible arguments.

        Please note however the default command data class, :class:`CommandData`
        does not perform path expansion.  So instead you have to handle this
        yourself with :meth:`map_path`.
        """
        raise NotImplementedError("`get_command_data` must be implemented")

    # TODO: finish map_path() implementation
    def map_path(self, path):
        """
        Takes a string argument.  Translates a given path for any OS to
        what it should be on this particular node.  This does not communicate
        with the master.
        """
        self._check_map_path_inputs(path)
        path = self.expandvars(path)
        return path

    def expandvars(self, value, environment=None, expand=None):
        """
        Expands variables inside of a string using an environment.  Exp

        :param string value:
            The path to expand

        :param dict environment:
            The environment to use for expanding ``value``.  If this
            value is None (the default) then we'll use :meth:`get_environment`
            to build this value.

        :param bool expand:
            When not provided we use the ``jobtype_expandvars`` configuration
            value to set the default.  When this value is True we'll
            perform environment variable expansion otherwise we return
            ``value`` untouched.
        """
        self._check_expandvars_inputs(value, environment)

        if expand is None:
            expand = config.get("jobtype_expandvars")

        if not expand:
            return value

        if environment is None:
            environment = self.get_environment()

        return Template(expanduser(value)).safe_substitute(**environment)

    def start(self):
        """
        This method is called when the job type should start
        working.  Depending on the job type's implementation this will
        prepare and start one more more processes.
        """
        environment = self.get_environment()
        command_data = self.get_command_data()

        if isinstance(command_data, self.COMMAND_DATA):
            command_data = [command_data]

        for command in command_data:
            if not isinstance(command, self.COMMAND_DATA):
                raise TypeError(
                    "Expected `%s` instances from "
                    "get_command_data()" % self.COMMAND_DATA.__name__)

            command.validate()
            command.set_default_environment(environment)
            self._spawn_process(command)

    def stop(self, signal="KILL"):
        """
        This method is called when the job type should stop
        running.  This will terminate any processes associated with
        this job type and also inform the master of any state changes
        to an associated task or tasks.

        :param string signal:
            The signal to send the any running processes.  Valid options
            are KILL, TERM or INT.
        """
        logger.debug("JobType.stop() called, signal: %s", signal)
        assert signal in ("KILL", "TERM", "INT")

        self.stop_called = True

        for _, process in self.processes.iteritems():
            if signal == "KILL":
                process.protocol.kill()
            elif signal == "TERM":
                process.protocol.terminate()
            elif signal == "INT":
                process.protocol.interrupt()
            else:
                raise NotImplementedError(
                    "Don't know how to handle signal %r" % signal)

        # TODO: notify master of stopped task(s)
        # TODO: chain this callback to the completion of our request to master

    def format_error(self, error):
        """
        Takes some kind of object, typically an instance of
        :class:`.Exception` or :class`.Failure` and produces a human readable
        string.  If we don't know how to format the request object an error
        will be logged and nothing will be returned
        """
        # It's likely that the process has terminated abnormally without
        # generating a trace back.  Modify the reason a little to better
        # reflect this.
        if isinstance(error, Failure):
            if error.type is ProcessTerminated and error.value.exitCode > 0:
                return "Process may have terminated abnormally, please check " \
                       "the logs.  Message from error " \
                       "was %r" % error.value.message

                # TODO: there are other types of errors from Twisted we should
                # format here

        elif isinstance(error, Exception):
            return str(error)

        elif isinstance(error, STRING_TYPES):
            return error

        elif error is None:
            logger.error("No error was defined for this failure.")

        else:
            logger.error("Don't know how to format %r as a string", error)

    # TODO: modify this function to support batch updates
    def set_states(self, tasks, state, error=None):
        """
        Wrapper around :meth:`set_state` that that allows you to
        the state on the master for multiple tasks at once.
        """
        self._check_set_states_inputs(tasks, state)

        for task in tasks:
            self.set_task_state(task, state, error=error)

    # TODO: refactor so `task` is an integer, not a dictionary
    def set_task_state(self, task, state, error=None):
        """
        Sets the state of the given task

        :param dict task:
            The dictionary containing the task we're changing the state
            for.

        :param string state:
            The state to change ``task`` to

        :type error: string, :class:`Exception`
        :param error:
            If the state is changing to 'error' then also set the
            ``last_error`` column.  Any exception instance that is
            passed to this keyword will be passed through
            :meth:`format_exception` first to format it.
        """
        if state not in WorkState:
            logger.error(
                "Cannot set state for task %r to %r, %r is an invalid "
                "state", task, state, state)

        elif not isinstance(task, dict):
            logger.error(
                "Expected a dictionary for `task`, cannot change state")

        elif "id" not in task or not isinstance(task["id"], INTEGER_TYPES):
            logger.error(
                "Expected to find 'id' in `task` or for `task['id']` to "
                "be an integer.")

        elif task not in self.assignment["tasks"]:
            logger.error(
                "Cannot set state, expected task %r to be a member of this "
                "job type's assignments", task)

        else:
            if state != WorkState.FAILED and error:
                logger.warning(
                    "Resetting `error` to None, state is not WorkState.FAILED")
                error = None

            # The task has failed
            if state == WorkState.FAILED:
                error = self.format_error(error)
                logger.error("Task %r failed: %r", task, error)
                if task["id"] not in self.failed_tasks:
                    self.failed_tasks.add(task["id"])

            # `error` shouldn't be set if the state is not a failure
            elif error is not None:
                logger.warning(
                    "`set_state` only allows `error` to be set if `state` is "
                    "'failed'.  Discarding error.")
                error = None

            url = "%s/jobs/%s/tasks/%s" % (
                config["master_api"], self.assignment["job"]["id"], task["id"])
            data = {"state": state}

            # If the error has been set then update the data we're
            # going to POST
            if isinstance(error, STRING_TYPES):
                data.update(last_error=error)

            elif error is not None:
                logger.error("Expected a string for `error`")

            def post_update(post_url, post_data, delay=0):
                post_func = partial(
                    post, post_url,
                    data=post_data,
                    callback=lambda x: result_callback(
                        post_url, post_data, task["id"], state, x),
                    errback=lambda x: error_callback(post_url, post_data, x))
                reactor.callLater(delay, post_func)

            def result_callback(cburl, cbdata, task_id, cbstate, response):
                if 500 <= response.code < 600:
                    logger.error(
                        "Error while posting state update for task %s "
                        "to %s, return code is %s, retrying",
                        task_id, cbstate, response.code)
                    post_update(cburl, cbdata, delay=http_retry_delay())

                elif response.code != OK:
                    # Nothing else we could do about that, this is
                    # a problem on our end.  We should only encounter
                    # this error during development
                    logger.error(
                        "Could not set state for task %s to %s, server "
                        "response code was %s",
                        task_id, cbstate, response.code)

                else:
                    logger.info(
                        "Set state of task %s to %s on master",
                        task_id, cbstate)
                    if cbstate == WorkState.DONE \
                            and task_id not in self.finished_tasks:
                        self.finished_tasks.add(task_id)

            def error_callback(cburl, cbdata, failure_reason):
                logger.error(
                    "Error while posting state update for task, %s, retrying",
                    failure_reason)
                post_update(cburl, cbdata, delay=http_retry_delay())

            # Initial attempt to make an update with an explicit zero
            # delay.
            post_update(url, data, delay=0)

            # TODO: if new state is the equiv. of 'stopped', stop the process
            # and POST the change

    def get_local_task_state(self, task_id):
        """
        Returns None if the state of this task has not been changed
        locally since this assignment has started.  This method does
        not communicate with the master.
        """
        if task_id in self.finished_tasks:
            return WorkState.DONE
        elif task_id in self.failed_tasks:
            return WorkState.FAILED
        else:
            return None

    def is_successful(self, reason):
        """
        **Overridable**.  This method that determines whether the process
        referred to by a protocol instance has exited successfully.

        The default implementation returns ``True`` if the process's return
        code was 0 and ``False` in all other cases.  If you need to
        modify this behavior please be aware that ``reason`` may be an
        integer or an instance of
        :class:`twisted.internet.error.ProcessTerminated` if the process
        terminated without errors or an instance of
        :class:`twisted.python.failure.Failure` if there were problems.

        :raises NotImplementedError:
            Raised if we encounter a condition that the base implementation
            is unable to handle.
        """
        if reason == 0:
            return True
        elif isinstance(reason, INTEGER_TYPES):
            return False
        elif hasattr(reason, "type"):
            return (
                reason.type is ProcessDone and
                reason.value.exitCode == 0)
        else:
            raise NotImplementedError(
                "Don't know how to handle is_successful(%r)" % reason)

    def before_start(self):
        """
        **Overridable**.  This method called directly before :meth:`start`
        itself is called.

        The default implementation does nothing and values returned from
        this method are ignored.
        """
        pass

    def before_spawn_process(self, command, protocol):
        """
        **Overridable**.  This method called directly before a process is
        spawned.

        By default this method does nothing except log information about
        the command we're about to launch both the the agent's log and to
        the log file on disk.

        :param CommandData command:
            An instance of :class:`CommandData` which contains the
            environment to use, command and arguments.  Modifications to
            this object will be applied to the process being spawned.

        :param ProcessProtocol protocol:
            An instance of :class:`pyfarm.jobtypes.core.process.ProcessProtocol`
            which contains the protocol used to communicate between the
            process and this job type.
        """
        logger.info("Spawning %r", command)
        logpool.log(protocol.uuid, STDERR,
                    "Command: %s" % command.command)
        logpool.log(protocol.uuid, STDERR,
                    "Arguments: %s" % (command.arguments, ))
        logpool.log(protocol.uuid, STDERR, "Work Dir: %s" % command.cwd)
        logpool.log(protocol.uuid, STDERR, "User/Group: %s %s" % (
            command.user, command.group))
        logpool.log(protocol.uuid, STDERR, "Environment:")
        logpool.log(protocol.uuid, STDERR, pformat(command.env, indent=4))

    def process_stopped(self, protocol, reason):
        """
        **Overridable**.  This method called when a child process stopped
        running.

        The default implementation will mark all tasks in the current
        assignment as done or failed of there was at least one failed process.
        """
        if len(self.failed_processes) == 0:
            for task in self.assignment["tasks"]:
                if task["id"] not in self.failed_tasks:
                    self.set_task_state(task, WorkState.DONE)
                else:
                    logger.info(
                        "Task %r is already in failed tasks, not setting state "
                        "to %s", task["id"], WorkState.DONE)
        else:
            for task in self.assignment["tasks"]:
                if task["id"] not in self.finished_tasks:
                    self.set_task_state(task, WorkState.FAILED)
                else:
                    logger.info(
                        "Task %r is already in finished tasks, not setting "
                        "state to %s", task["id"], WorkState.FAILED)

    def process_started(self, protocol):
        """
        **Overridable**.  This method is called when a child process started
        running.

        The default implementation will mark all tasks in the current
        assignment as running.
        """
        for task in self.assignment["tasks"]:
            self.set_task_state(task, WorkState.RUNNING)

    def process_output(self, protocol, output, line_fragments, line_handler):
        """
        This is a mid-level method which takes output from a process protocol
        then splits and processes it to ensure we pass complete output lines
        to the other methods.

        Implementors who wish to process the output line by line should
        override :meth:`preprocess_stdout_line`, :meth:`preprocess_stdout_line`,
        :meth:`process_stdout_line` or :meth:`process_stderr_line` instead.
        This method is a glue method between other parts of the job type and
        should only be overridden if there's a problem or you want to change
        how lines are split.

        :type protocol: :class:`.ProcessProtocol`
        :param protocol:
            The protocol instance which produced ``output``

        :param string output:
            The blob of text or line produced

        :param dict line_fragments:
            The line fragment dictionary containing individual line
            fragments.  This will be either ``self._stdout_line_fragments`` or
            ``self._stderr_line_fragments``.

        :param callable line_handler:
            The function to handle any lines produced.  This will be either
            :meth:`handle_stdout_line` or :meth:`handle_stderr_line`

        :return:
            This method returns nothing by default and any return value
            produced by this method will not be consumed by other methods.
        """
        dangling_fragment = None

        if "\n" in output:
            ends_on_fragment = True

            if output[-1] == "\n":
                ends_on_fragment = False

            lines = output.split("\n")
            if ends_on_fragment:
                dangling_fragment = lines.pop(-1)

            for line in lines:
                if protocol.uuid in line_fragments:
                    line_out = line_fragments.pop(protocol.uuid)
                    line_out += line
                    if line_out and line_out[-1] == "\r":
                        line_out = line_out[:-1]

                    line_handler(protocol, line_out)

                else:
                    if line and line[-1] == "\r":
                        line = line[:-1]

                    line_handler(protocol, line)

            if ends_on_fragment:
                line_fragments[protocol.uuid] = dangling_fragment
        else:
            if protocol.uuid in line_fragments:
                line_fragments[protocol.uuid] += output

            else:
                line_fragments[protocol.uuid] = output

    def handle_stdout_line(self, protocol, stdout):
        """
        Takes a :class:`.ProcessProtocol` instance and ``stdout``
        line produced by :meth:`process_output` and runs it through all
        the steps necessary to preprocess, format, log and handle the line.

        The default implementation will run ``stdout`` through several methods
        in order:

            * :meth:`preprocess_stdout_line`
            * :meth:`format_stdout_line`
            * :meth:`log_stdout_line`
            * :meth:`process_stdout_line`

        .. warning::

            This method is not private however it's advisable to override
            the methods above instead of this one.  Unlike this method,
            which is more generalized and invokes several other methods,
            the above provide more targeted functionality.

        :type protocol: :class:`.ProcessProtocol`
        :param protocol:
            The protocol instance which produced ``stdout``

        :param string stderr:
            A complete line to ``stderr`` being emitted by the process

        :return:
            This method returns nothing by default and any return value
            produced by this method will not be consumed by other methods.
        """
        # Preprocess
        preprocessed = self.preprocess_stdout_line(protocol, stdout)
        if preprocessed is not None:
            stdout = preprocessed
        if preprocessed is False:
            return

        # Format
        formatted = self.format_stdout_line(protocol, stdout)
        if formatted is not None:
            stdout = formatted

        # Log
        self.log_stdout_line(protocol, stdout)

        # Custom Processing
        self.process_stdout_line(protocol, stdout)

    def handle_stderr_line(self, protocol, stderr):
        """
        **Overridable**.  Takes a :class:`.ProcessProtocol` instance and
        ``stderr`` produced by :meth:`process_output` and runs it through all
        the steps necessary to preprocess, format, log and handle the line.

        The default implementation will run ``stderr`` through several methods
        in order:

            * :meth:`preprocess_stderr_line`
            * :meth:`format_stderr_line`
            * :meth:`log_stderr_line`
            * :meth:`process_stderr_line`

        .. warning::

            This method is overridable however it's advisable to override
            the methods above instead.  Unlike this method, which is more
            generalized and invokes several other methods, the above provide
            more targeted functionality.

        :type protocol: :class:`.ProcessProtocol`
        :param protocol:
            The protocol instance which produced ``stdout``

        :param string stderr:
            A complete line to ``stderr`` being emitted by the process

        :return:
            This method returns nothing by default and any return value
            produced by this method will not be consumed by other methods.
        """
        # Preprocess
        preprocessed = self.preprocess_stderr_line(protocol, stderr)
        if preprocessed is not None:
            stderr = preprocessed
        if preprocessed == False:
            return

        # Format
        formatted = self.format_stderr_line(protocol, stderr)
        if formatted is not None:
            stderr = formatted

        # Log
        self.log_stderr_line(protocol, stderr)

        # Custom Processing
        self.process_stderr_line(protocol, stderr)

    def preprocess_stdout_line(self, protocol, stdout):
        """
        **Overridable**.  Provides the ability to manipulate ``stdout`` or
        ``protocol`` before it's passed into any other line handling methods.

        *The default implementation does nothing.*

        :type protocol: :class:`.ProcessProtocol`
        :param protocol:
            The protocol instance which produced ``stdout``

        :param string stderr:
            A complete line to ``stdout`` before any formatting or logging
            has occurred.

        :rtype: string
        :return:
            This method returns nothing by default but when overridden should
            return a string which will be used in line handling methods such
            as :meth:`format_stdout_line`, :meth:`log_stdout_line` and
            :meth:`process_stdout_line`.
        """
        pass

    def preprocess_stderr_line(self, protocol, stderr):
        """
        **Overridable**.  Formats a line from ``stdout`` before it's passed onto
        methods such as :meth:`log_stdout_line` and :meth:`process_stdout_line`.

        *The default implementation does nothing.*

        :type protocol: :class:`.ProcessProtocol`
        :param protocol:
            The protocol instance which produced ``stderr``

        :param string stderr:
            A complete line to ``stderr`` before any formatting or logging
            has occurred.

        :rtype: string
        :return:
            This method returns nothing by default but when overridden should
            return a string which will be used in line handling methods such
            as :meth:`format_stderr_line`, :meth:`log_stderr_line` and
            :meth:`process_stderr_line`.
        """
        pass

    def format_stdout_line(self, protocol, stdout):
        """
        **Overridable**.  Formats a line from ``stdout`` before it's passed onto
        methods such as :meth:`log_stdout_line` and :meth:`process_stdout_line`.

        *The default implementation does nothing.*

        :type protocol: :class:`.ProcessProtocol`
        :param protocol:
            The protocol instance which produced ``stdout``

        :param string stdout:
            A complete line from process to format and return.

        :rtype: string
        :return:
            This method returns nothing by default but when overridden should
            return a string which will be used in :meth:`log_stdout_line` and
            :meth:`process_stdout_line`
        """
        pass

    def format_stderr_line(self, protocol, stderr):
        """
        **Overridable**.  Formats a line from ``stderr`` before it's passed onto
        methods such as :meth:`log_stderr_line` and :meth:`process_stderr_line`.

        *The default implementation does nothing.*

        :type protocol: :class:`.ProcessProtocol`
        :param protocol:
            The protocol instance which produced ``stderr``

        :param string stderr:
            A complete line from the process to format and return.

        :rtype: string
        :return:
            This method returns nothing by default but when overridden should
            return a string which will be used in :meth:`log_stderr_line` and
            :meth:`process_stderr_line`
        """
        pass

    def log_stdout_line(self, protocol, stdout):
        """
        **Overridable**.  Called when we receive a complete line on stdout from
        the process.

        The default implementation will use the global logging pool to
        log ``stdout`` to a file.

        :type protocol: :class:`.ProcessProtocol`
        :param protocol:
            The protocol instance which produced ``stdout``

        :param string stderr:
            A complete line to ``stdout`` that has been formatted and is ready
            to log to a file.

        :return:
            This method returns nothing by default and any return value
            produced by this method will not be consumed by other methods.
        """
        if config["jobtype_capture_process_output"]:
            process_stdout.info("task %r: %s", protocol.id, stdout)
        else:
            logpool.log(protocol.uuid, STDOUT, stdout)

    def log_stderr_line(self, protocol, stderr):
        """
        **Overridable**.  Called when we receive a complete line on stderr from
        the process.

        The default implementation will use the global logging pool to
        log ``stderr`` to a file.

        :type protocol: :class:`.ProcessProtocol`
        :param protocol:
            The protocol instance which produced ``stderr``

        :param string stderr:
            A complete line to ``stderr`` that has been formatted and is ready
            to log to a file.

        :return:
            This method returns nothing by default and any return value
            produced by this method will not be consumed by other methods.
        """
        if config["jobtype_capture_process_output"]:
            process_stdout.info("task %r: %s", protocol.id, stderr)
        else:
            logpool.log(protocol.uuid, STDERR, stderr)

    def process_stderr_line(self, protocol, stderr):
        """
        **Overridable**.  This method is called when we receive a complete
        line to ``stderr``.  The line will be preformatted and will already
        have been sent for logging.

        *The default implementation sends ``stderr`` and ``protocol`` to
        :meth:`process_stdout_line`.*

        :type protocol: :class:`.ProcessProtocol`
        :param protocol:
            The protocol instance which produced ``stderr``

        :param string stderr:
            A complete line to ``stderr`` after it has been formatted and
            logged.

        :return:
            This method returns nothing by default and any return value
            produced by this method will not be consumed by other methods.
        """
        self.process_stdout_line(protocol, stderr)

    def process_stdout_line(self, protocol, stdout):
        """
        **Overridable**.  This method is called when we receive a complete
        line to ``stdout``.  The line will be preformatted and will already
        have been sent for logging.

        *The default implementation does nothing.*

        :type protocol: :class:`.ProcessProtocol`
        :param protocol:
            The protocol instance which produced ``stderr``

        :param string stderr:
            A complete line to ``stdout`` after it has been formatted and
            logged.

        :return:
            This method returns nothing by default and any return value
            produced by this method will not be consumed by other methods.
        """
        pass
